import re

from django.conf import settings
from notifications.settings import CONFIG_DEFAULTS

CONFIG_DEFAULTS.update({'USE_JSONFIELD': True})

HOST = getattr(settings, 'OPENWISP_NOTIFICATIONS_HOST', None)

CACHE_TIMEOUT = getattr(
    settings, 'OPENWISP_NOTIFICATIONS_CACHE_TIMEOUT', 2 * 24 * 60 * 60
)

IGNORE_ENABLED_ADMIN = getattr(
    settings, 'OPENWISP_NOTIFICATIONS_IGNORE_ENABLED_ADMIN', []
)
POPULATE_PREFERENCES_ON_MIGRATE = getattr(
    settings, 'OPENWISP_NOTIFICATIONS_POPULATE_PREFERENCES_ON_MIGRATE', True
)
NOTIFICATION_STORM_PREVENTION = getattr(
    settings,
    'OPENWISP_NOTIFICATIONS_NOTIFICATION_STORM_PREVENTION',
    {
        'short_term_time_period': 10,
        'short_term_notification_count': 6,
        'long_term_time_period': 180,
        'long_term_notification_count': 30,
        'initial_backoff': 1,
        'backoff_increment': 1,
        'max_allowed_backoff': 15,
    },
)

SOUND = getattr(
    settings,
    'OPENWISP_NOTIFICATIONS_SOUND',
    'openwisp-notifications/audio/notification_bell.mp3',
)
# Remove the leading "/static/" here as it will
# conflict with the "static()" call in context_processors.py.
# This is done for backward compatibility.
SOUND = re.sub('^/static/', '', SOUND)


def get_config():
    user_config = getattr(settings, 'OPENWISP_NOTIFICATIONS_CONFIG', {})
    config = CONFIG_DEFAULTS.copy()
    config.update(user_config)
    return config



from unittest.mock import patch

from django.db.utils import OperationalError
from django.test import TestCase
from django.utils import timezone

from openwisp_notifications import tasks
from openwisp_notifications.signals import notify
from openwisp_notifications.swapper import load_model
from openwisp_users.tests.utils import TestOrganizationMixin

from ..types import get_notification_configuration
from .test_helpers import mock_notification_types, register_notification_type

IgnoreObjectNotification = load_model('IgnoreObjectNotification')
Notification = load_model('Notification')
on_queryset = IgnoreObjectNotification.objects


class TestIgnoreObjectNotification(TestOrganizationMixin, TestCase):
    def setUp(self):
        self.obj = self._get_org_user()
        self.admin = self._get_admin()

    def test_object_notification(self):
        IgnoreObjectNotification.objects.create(
            object=self.obj, user=self.admin, valid_till=timezone.now()
        )
        # Celery task deletes it right away
        self.assertEqual(on_queryset.count(), 0)

    @patch('openwisp_notifications.tasks.delete_ignore_object_notification.apply_async')
    def test_delete_object_busy_worker(self, mocked_task):
        IgnoreObjectNotification.objects.create(
            object=self.obj, user=self.admin, valid_till=timezone.now()
        )
        self.assertEqual(on_queryset.count(), 1)

    @patch('openwisp_notifications.tasks.delete_ignore_object_notification.apply_async')
    def test_notification_for_disabled_object(self, mocked_task):
        IgnoreObjectNotification.objects.create(
            object=self.obj,
            user=self.admin,
            valid_till=(timezone.now() + timezone.timedelta(days=1)),
        )
        notify.send(sender=self.admin, type='default', target=self.obj)
        self.assertEqual(Notification.objects.count(), 0)

    @mock_notification_types
    @patch('openwisp_notifications.tasks.delete_ignore_object_notification.apply_async')
    def test_related_object_deleted(self, *args):
        type_config = get_notification_configuration('default')
        register_notification_type('test', type_config, models=[self.obj._meta.model])
        IgnoreObjectNotification.objects.create(
            object=self.obj,
            user=self.admin,
            valid_till=(timezone.now() + timezone.timedelta(days=1)),
        )
        self.obj.delete()
        self.assertEqual(on_queryset.count(), 0)

    @patch.object(
        IgnoreObjectNotification.objects, 'filter', side_effect=OperationalError
    )
    def test_related_object_deleted_task(self, *args):
        tasks.delete_obsolete_objects.delay(
            self.obj._meta.app_label, self.obj._meta.model_name, self.obj.id
        )



import logging

from django.db import models
from rest_framework import serializers
from rest_framework.exceptions import NotFound

from openwisp_notifications.exceptions import NotificationRenderException
from openwisp_notifications.swapper import load_model

logger = logging.getLogger(__name__)

Notification = load_model('Notification')
NotificationSetting = load_model('NotificationSetting')
IgnoreObjectNotification = load_model('IgnoreObjectNotification')


class ContentTypeField(serializers.Field):
    def to_representation(self, obj):
        return obj.model


class CustomListSerializer(serializers.ListSerializer):
    def to_representation(self, data):
        iterable = data.all() if isinstance(data, models.Manager) else data
        data_list = []
        for item in iterable:
            try:
                data_list.append(self.child.to_representation(item))
            except NotificationRenderException as e:
                logger.error(e)
        return data_list


class NotificationSerializer(serializers.ModelSerializer):
    actor_content_type = ContentTypeField(read_only=True)
    target_content_type = ContentTypeField(read_only=True)
    action_object_content_type = ContentTypeField(read_only=True)

    class Meta:
        model = Notification
        exclude = ['description', 'deleted', 'public']
        extra_fields = ['message', 'email_subject', 'target_url']

    def get_field_names(self, declared_fields, info):
        model_fields = super().get_field_names(declared_fields, info)
        return model_fields + self.Meta.extra_fields

    @property
    def data(self):
        try:
            return super().data
        except NotificationRenderException as e:
            logger.error(e)
            raise NotFound


class NotificationListSerializer(NotificationSerializer):
    description = serializers.CharField(source='rendered_description')

    class Meta(NotificationSerializer.Meta):
        fields = [
            'id',
            'message',
            'description',
            'unread',
            'target_url',
            'email_subject',
            'timestamp',
            'level',
        ]
        exclude = None
        list_serializer_class = CustomListSerializer


class NotificationSettingSerializer(serializers.ModelSerializer):
    class Meta:
        model = NotificationSetting
        exclude = ['user']
        read_only_fields = ['organization', 'type']


class IgnoreObjectNotificationSerializer(serializers.ModelSerializer):
    class Meta:
        model = IgnoreObjectNotification
        exclude = ['user']
        read_only_fields = [
            'object_content_type',
            'object_id',
        ]



from openwisp_notifications.swapper import load_model
from openwisp_users.api.filters import OrganizationMembershipFilter

NotificationSetting = load_model('NotificationSetting')


class NotificationSettingFilter(OrganizationMembershipFilter):
    class Meta(OrganizationMembershipFilter.Meta):
        model = NotificationSetting
        fields = OrganizationMembershipFilter.Meta.fields + ['type']



import logging

from celery.exceptions import OperationalError
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.core.cache import cache
from django.db import transaction
from django.db.models import Q
from django.db.models.query import QuerySet
from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver
from django.utils import timezone
from django.utils.translation import gettext as _

from openwisp_notifications import settings as app_settings
from openwisp_notifications import tasks
from openwisp_notifications.exceptions import NotificationRenderException
from openwisp_notifications.swapper import load_model, swapper_load_model
from openwisp_notifications.types import (
    NOTIFICATION_ASSOCIATED_MODELS,
    get_notification_configuration,
)
from openwisp_notifications.websockets import handlers as ws_handlers
from openwisp_utils.admin_theme.email import send_email

logger = logging.getLogger(__name__)

EXTRA_DATA = app_settings.get_config()['USE_JSONFIELD']

User = get_user_model()

Notification = load_model('Notification')
NotificationSetting = load_model('NotificationSetting')
IgnoreObjectNotification = load_model('IgnoreObjectNotification')

Group = swapper_load_model('openwisp_users', 'Group')
OrganizationUser = swapper_load_model('openwisp_users', 'OrganizationUser')
Organization = swapper_load_model('openwisp_users', 'Organization')


def notify_handler(**kwargs):
    """
    Handler function to create Notification instance upon action signal call.
    """
    # Pull the options out of kwargs
    kwargs.pop('signal', None)
    actor = kwargs.pop('sender')
    public = bool(kwargs.pop('public', True))
    description = kwargs.pop('description', None)
    timestamp = kwargs.pop('timestamp', timezone.now())
    recipient = kwargs.pop('recipient', None)
    notification_type = kwargs.pop('type', None)
    target = kwargs.get('target', None)
    target_org = getattr(target, 'organization_id', None)
    try:
        notification_template = get_notification_configuration(notification_type)
    except NotificationRenderException as error:
        logger.error(f'Error encountered while creating notification: {error}')
        return
    level = kwargs.pop(
        'level', notification_template.get('level', Notification.LEVELS.info)
    )
    verb = notification_template.get('verb', kwargs.pop('verb', None))
    user_app_name = User._meta.app_label

    where = Q(is_superuser=True)
    not_where = Q()
    where_group = Q()
    if target_org:
        org_admin_query = Q(
            **{
                f'{user_app_name}_organizationuser__organization': target_org,
                f'{user_app_name}_organizationuser__is_admin': True,
            }
        )
        where = where | (Q(is_staff=True) & org_admin_query)
        where_group = org_admin_query

        # We can only find notification setting if notification type and
        # target organization is present.
        if notification_type:
            # Create notification for users who have opted for receiving notifications.
            # For users who have not configured web_notifications,
            # use default from notification type
            web_notification = Q(notificationsetting__web=True)
            if notification_template['web_notification']:
                web_notification |= Q(notificationsetting__web=None)

            notification_setting = web_notification & Q(
                notificationsetting__type=notification_type,
                notificationsetting__organization_id=target_org,
                notificationsetting__deleted=False,
            )
            where = where & notification_setting
            where_group = where_group & notification_setting

    # Ensure notifications are only sent to active user
    where = where & Q(is_active=True)
    where_group = where_group & Q(is_active=True)

    # We can only find ignore notification setting if target object is present
    if target:
        not_where = Q(
            ignoreobjectnotification__object_id=target.pk,
            ignoreobjectnotification__object_content_type=ContentType.objects.get_for_model(
                target._meta.model
            ),
        ) & (
            Q(ignoreobjectnotification__valid_till=None)
            | Q(ignoreobjectnotification__valid_till__gt=timezone.now())
        )

    if recipient:
        # Check if recipient is User, Group or QuerySet
        if isinstance(recipient, Group):
            recipients = recipient.user_set.filter(where_group)
        elif isinstance(recipient, QuerySet):
            recipients = recipient.distinct()
        elif isinstance(recipient, list):
            recipients = recipient
        else:
            recipients = [recipient]
    else:
        recipients = (
            User.objects.prefetch_related(
                'notificationsetting_set', 'ignoreobjectnotification_set'
            )
            .order_by('date_joined')
            .filter(where)
            .exclude(not_where)
            .distinct()
        )
    optional_objs = [
        (kwargs.pop(opt, None), opt) for opt in ('target', 'action_object')
    ]

    notification_list = []
    for recipient in recipients:
        notification = Notification(
            recipient=recipient,
            actor=actor,
            verb=str(verb),
            public=public,
            description=description,
            timestamp=timestamp,
            level=level,
            type=notification_type,
        )

        # Set optional objects
        for obj, opt in optional_objs:
            if obj is not None:
                setattr(notification, '%s_object_id' % opt, obj.pk)
                setattr(
                    notification,
                    '%s_content_type' % opt,
                    ContentType.objects.get_for_model(obj),
                )
        if kwargs and EXTRA_DATA:
            notification.data = kwargs
        notification.save()
        notification_list.append(notification)

    return notification_list


@receiver(post_save, sender=Notification, dispatch_uid='send_email_notification')
def send_email_notification(sender, instance, created, **kwargs):
    # Abort if a new notification is not created
    if not created:
        return
    # Get email preference of user for this type of notification.
    target_org = getattr(getattr(instance, 'target', None), 'organization_id', None)
    if instance.type and target_org:
        try:
            notification_setting = instance.recipient.notificationsetting_set.get(
                organization=target_org, type=instance.type
            )
        except NotificationSetting.DoesNotExist:
            return
        email_preference = notification_setting.email_notification
    else:
        # We can not check email preference if notification type is absent,
        # or if target_org is not present
        # therefore send email anyway.
        email_preference = True

    email_verified = instance.recipient.emailaddress_set.filter(
        verified=True, email=instance.recipient.email
    ).exists()

    if not (email_preference and instance.recipient.email and email_verified):
        return

    try:
        subject = instance.email_subject
    except NotificationRenderException:
        # Do not send email if notification is malformed.
        return
    url = instance.data.get('url', '') if instance.data else None
    body_text = instance.email_message
    if url:
        target_url = url
    elif instance.target:
        target_url = instance.redirect_view_url
    else:
        target_url = None
    if target_url:
        body_text += _('\n\nFor more information see %(target_url)s.') % {
            'target_url': target_url
        }

    send_email(
        subject=subject,
        body_text=body_text,
        body_html=instance.email_message,
        recipients=[instance.recipient.email],
        extra_context={
            'call_to_action_url': target_url,
            'call_to_action_text': _('Find out more'),
        },
    )

    # flag as emailed
    instance.emailed = True
    # bulk_update is used to prevent emitting post_save signal
    Notification.objects.bulk_update([instance], ['emailed'])


@receiver(post_save, sender=Notification, dispatch_uid='clear_notification_cache_saved')
@receiver(
    post_delete, sender=Notification, dispatch_uid='clear_notification_cache_deleted'
)
def clear_notification_cache(sender, instance, **kwargs):
    Notification.invalidate_unread_cache(instance.recipient)
    # Reload notification widget only if notification is created or deleted
    # Display notification toast when a new notification is created
    ws_handlers.notification_update_handler(
        recipient=instance.recipient,
        reload_widget=kwargs.get('created', True),
        notification=instance if kwargs.get('created', None) else None,
    )


@receiver(post_delete, dispatch_uid='delete_obsolete_objects')
def related_object_deleted(sender, instance, **kwargs):
    """
    Delete Notification and IgnoreObjectNotification objects having
    "instance" as related object.
    """
    if sender not in NOTIFICATION_ASSOCIATED_MODELS:
        return
    instance_id = getattr(instance, 'pk', None)
    if instance_id:
        instance_model = instance._meta.model_name
        instance_app_label = instance._meta.app_label
        tasks.delete_obsolete_objects.delay(
            instance_app_label, instance_model, instance_id
        )


def notification_type_registered_unregistered_handler(sender, **kwargs):
    try:
        tasks.ns_register_unregister_notification_type.delay()
    except OperationalError:
        logger.warn(
            '\tCelery broker is unreachable, skipping populating data for user(s) '
            'notification preference(s).\n'
            '\tMake sure that celery broker is running and reachable by celery workers.\n'
            '\tYou can use following command later '
            'to populate data for user(s) notification preference(s).\n\n'
            '\t\t python manage.py populate_notification_preferences\n'
        )


@receiver(
    post_save,
    sender=OrganizationUser,
    dispatch_uid='create_orguser_notification_setting',
)
def organization_user_post_save(instance, created, **kwargs):
    transaction.on_commit(
        lambda: tasks.update_org_user_notificationsetting.delay(
            org_user_id=instance.pk,
            user_id=instance.user_id,
            org_id=instance.organization_id,
            is_org_admin=instance.is_admin,
        )
    )


@receiver(
    post_delete,
    sender=OrganizationUser,
    dispatch_uid='delete_orguser_notification_setting',
)
def notification_setting_delete_org_user(instance, **kwargs):
    tasks.ns_organization_user_deleted.delay(
        user_id=instance.user_id, org_id=instance.organization_id
    )


@receiver(post_save, sender=User, dispatch_uid='user_notification_setting')
def update_superuser_notification_settings(instance, created, **kwargs):
    transaction.on_commit(
        lambda: tasks.update_superuser_notification_settings.delay(
            instance.pk, instance.is_superuser, created
        )
    )


@receiver(
    post_save, sender=Organization, dispatch_uid='org_created_notification_setting'
)
def notification_setting_org_created(created, instance, **kwargs):
    if created:
        transaction.on_commit(lambda: tasks.ns_organization_created.delay(instance.pk))


@receiver(
    post_save,
    sender=IgnoreObjectNotification,
    dispatch_uid='schedule_object_notification_deletion',
)
def schedule_object_notification_deletion(instance, created, **kwargs):
    if instance.valid_till is not None:
        tasks.delete_ignore_object_notification.apply_async(
            (instance.pk,), eta=instance.valid_till
        )


def register_notification_cache_update(model, signal, dispatch_uid=None):
    signal.connect(
        update_notification_cache,
        sender=model,
        dispatch_uid=dispatch_uid,
    )


def update_notification_cache(sender, instance, **kwargs):
    def invalidate_cache():
        content_type = ContentType.objects.get_for_model(instance)
        cache_key = Notification._cache_key(content_type.id, instance.id)
        cache.delete(cache_key)

    # execute cache invalidation only after changes have been committed to the DB
    transaction.on_commit(invalidate_cache)



from django.contrib import admin

from openwisp_notifications.base.admin import NotificationSettingAdminMixin
from openwisp_notifications.swapper import load_model
from openwisp_notifications.widgets import _add_object_notification_widget
from openwisp_users.admin import UserAdmin
from openwisp_utils.admin import AlwaysHasChangedMixin

Notification = load_model('Notification')
NotificationSetting = load_model('NotificationSetting')


class NotificationSettingInline(
    NotificationSettingAdminMixin, AlwaysHasChangedMixin, admin.TabularInline
):
    model = NotificationSetting
    extra = 0

    def has_change_permission(self, request, obj=None):
        return request.user.is_superuser or request.user == obj


UserAdmin.inlines = [NotificationSettingInline] + UserAdmin.inlines

_add_object_notification_widget()



# chat/routing.py
from django.urls import path

from . import consumers as ow_consumers


def get_routes(consumer=None):
    if not consumer:
        consumer = ow_consumers
    return [
        path('ws/notification/', consumer.NotificationConsumer.as_asgi()),
    ]



from django.core.exceptions import ImproperlyConfigured
from django.template.loader import get_template

from .exceptions import NotificationRenderException

NOTIFICATION_TYPES = {
    'default': {
        'level': 'info',
        'verb': 'default verb',
        'verbose_name': 'Default Type',
        'email_subject': '[{site.name}] Default Notification Subject',
        'message': (
            'Default notification with {notification.verb} and level {notification.level}'
            ' by [{notification.target}]({notification.target_link})'
        ),
        'message_template': 'openwisp_notifications/default_message.md',
        'email_notification': True,
        'web_notification': True,
    },
    'generic_message': {
        'level': 'info',
        'verb': 'generic verb',
        'verbose_name': 'Generic Type',
        'email_subject': '[{site.name}] Generic Notification Subject',
        'message': (
            'Generic notification with {notification.verb} and level {notification.level}'
            ' by [{notification.actor}]({notification.actor_link})'
        ),
        'description': '{notification.description}',
        'email_notification': False,
        'web_notification': True,
    },
}

NOTIFICATION_CHOICES = [
    ('default', 'Default Type'),
    ('generic_message', 'Generic Message Type'),
]
NOTIFICATION_ASSOCIATED_MODELS = set()


def get_notification_configuration(notification_type):
    if not notification_type:
        return {}
    try:
        return NOTIFICATION_TYPES[notification_type]
    except KeyError:
        raise NotificationRenderException(
            f'No such Notification Type, {notification_type}'
        )


def _validate_notification_type(type_config):
    options = type_config.keys()
    assert 'level' in options
    assert 'verb' in options
    assert 'email_subject' in options
    assert ('message' in options) or ('message_template' in options)

    if 'message_template' in options:
        get_template(type_config['message_template'])

    if 'email_notification' not in options:
        type_config['email_notification'] = True

    if 'web_notification' not in options:
        type_config['web_notification'] = True

    return type_config


def register_notification_type(type_name, type_config, models=[]):
    """
    Registers a new notification type.
    """
    if not isinstance(type_name, str):
        raise ImproperlyConfigured('Notification Type name should be type `str`.')
    if not isinstance(type_config, dict):
        raise ImproperlyConfigured(
            'Notification Type configuration should be type `dict`.'
        )
    if type_name in NOTIFICATION_TYPES:
        raise ImproperlyConfigured(
            f'{type_name} is an already registered Notification Type.'
        )

    validated_type_config = _validate_notification_type(type_config)
    NOTIFICATION_TYPES.update({type_name: validated_type_config})
    _register_notification_choice(type_name, validated_type_config)
    NOTIFICATION_ASSOCIATED_MODELS.update(models)


def unregister_notification_type(type_name):
    if not isinstance(type_name, str):
        raise ImproperlyConfigured('Notification Type name should be type `str`')
    if type_name not in NOTIFICATION_TYPES:
        raise ImproperlyConfigured(f'No such Notification Type, {type_name}')

    NOTIFICATION_TYPES.pop(type_name)
    _unregister_notification_choice(type_name)


def _register_notification_choice(type_name, type_config):
    name = type_config.get('verbose_name', type_name)
    NOTIFICATION_CHOICES.append((type_name, name))


def _unregister_notification_choice(notification_type):
    for index, (key, name) in enumerate(NOTIFICATION_CHOICES):
        if key == notification_type:
            NOTIFICATION_CHOICES.pop(index)
            return
    raise ImproperlyConfigured(f'No such Notification Choice {notification_type}')



from django.dispatch import Signal

notify = Signal()
notify.__doc__ = """
Creates notification(s).

Sends arguments: 'recipient', 'actor', 'verb', 'action_object',
    'target', 'description', 'timestamp', 'level', 'type', etc.
"""



from swapper import load_model as swapper_load_model

from openwisp_notifications.apps import OpenwispNotificationsConfig as AppConfig


def load_model(model):
    return swapper_load_model(AppConfig.label, model)



from django.templatetags.static import static

from openwisp_notifications import settings as app_settings


def notification_api_settings(request):
    return {
        'OPENWISP_NOTIFICATIONS_HOST': app_settings.HOST,
        'OPENWISP_NOTIFICATIONS_SOUND': static(app_settings.SOUND),
    }



from django.forms import ModelForm, widgets

from openwisp_notifications.swapper import swapper_load_model


class NotificationSettingForm(ModelForm):
    def __init__(self, *args, **kwargs):
        instance = kwargs.get('instance')
        if instance:
            kwargs['initial'] = {
                'web': instance.web_notification,
                'email': (
                    instance.email_notification
                    if instance.web_notification
                    else instance.web_notification
                ),
            }
        super().__init__(*args, **kwargs)
        try:
            self.fields['organization'].choices = self.get_organization_choices()
        except KeyError:
            pass

    @classmethod
    def get_organization_choices(cls):
        if not hasattr(cls, 'organization_choices'):
            Organization = swapper_load_model('openwisp_users', 'organization')
            cls.organization_choices = [(None, '---------')] + list(
                Organization.objects.all().values_list('pk', 'name')
            )
        return cls.organization_choices

    class Meta:
        widgets = {
            'web': widgets.CheckboxInput,
            'email': widgets.CheckboxInput,
        }



from io import StringIO
from unittest.mock import patch

import swapper
from django.core import management
from django.test import TestCase

from openwisp_notifications import checks
from openwisp_notifications import settings as app_settings
from openwisp_notifications.swapper import load_model
from openwisp_users.tests.utils import TestOrganizationMixin

Notification = load_model('Notification')
NotificationSetting = load_model('NotificationSetting')
Organization = swapper.load_model('openwisp_users', 'Organization')


class TestManagementCommands(TestCase, TestOrganizationMixin):
    def test_create_notification_command(self):
        admin = self._get_admin()
        default_org = Organization.objects.first()
        ns = NotificationSetting(user=admin, organization=default_org, type='default')
        ns.save()
        management.call_command('create_notification')
        self.assertEqual(Notification.objects.count(), 1)
        n = Notification.objects.first()
        self.assertEqual(n.type, 'default')
        self.assertEqual(n.actor, default_org)
        self.assertEqual(n.recipient, admin)

    def test_create_notification_command_when_notificationsetting_disable(self):
        admin = self._get_admin()
        default_org = Organization.objects.first()
        ns = NotificationSetting(
            user=admin, organization=default_org, type='default', web=False, email=False
        )
        ns.save()
        management.call_command('create_notification')
        self.assertEqual(Notification.objects.count(), 0)

    @patch(
        'openwisp_notifications.tasks.ns_register_unregister_notification_type.delay'
    )
    def test_populate_notification_preferences_command(self, mocked_task):
        management.call_command('populate_notification_preferences')
        mocked_task.assert_called_once()


class TestChecks(TestCase, TestOrganizationMixin):
    @patch.object(
        app_settings,
        'HOST',
        'https://example.com',
    )
    def test_cors_not_configured(self):
        # If INSTALLED_APPS not configured
        with patch.multiple(
            'openwisp_notifications.types',
            NOTIFICATION_TYPES={},
            NOTIFICATION_CHOICES=[],
        ), patch('openwisp_utils.admin_theme.menu.MENU', {}), self.modify_settings(
            INSTALLED_APPS={'remove': 'corsheaders'}
        ), StringIO() as stderr:
            management.call_command('check', stderr=stderr)
            self.assertIn('django-cors-headers', stderr.getvalue())

        # If MIDDLEWARE not configured
        with patch.multiple(
            'openwisp_notifications.types',
            NOTIFICATION_TYPES={},
            NOTIFICATION_CHOICES=[],
        ), self.modify_settings(
            MIDDLEWARE={'remove': 'corsheaders.middleware.CorsMiddleware'}
        ), StringIO() as stderr:
            management.call_command('check', stderr=stderr)
            self.assertIn('django-cors-headers', stderr.getvalue())

    def test_ow_object_notification_setting_improperly_configured(self):
        def run_check():
            return checks.check_ow_object_notification_widget_setting(None).pop()

        with self.subTest('Test setting is not a list'):
            with patch.object(app_settings, 'IGNORE_ENABLED_ADMIN', tuple()):
                error_message = (
                    '"OPENWISP_NOTIFICATIONS_IGNORE_ENABLED_ADMIN" should be a list'
                )
                error = run_check()
                self.assertIn(error_message, error.hint)

        with self.subTest('Test setting does not contains dotted path string'):
            with patch.object(app_settings, 'IGNORE_ENABLED_ADMIN', [0]):
                error_message = (
                    '"OPENWISP_NOTIFICATIONS_IGNORE_ENABLED_ADMIN" should contain '
                    'dotted path string to ModelAdmin'
                )
                error = run_check()
                self.assertIn(error_message, error.hint)

        with self.subTest('Test setting dotted path is invalid'):
            path = 'openwisp_notifications.admin.DeviceAdmin'
            with patch.object(app_settings, 'IGNORE_ENABLED_ADMIN', [path]):
                error_message = (
                    f'Failed to import "{path}" defined in '
                    '"OPENWISP_NOTIFICATIONS_IGNORE_ENABLED_ADMIN".'
                )
                error = run_check()
                self.assertIn(error_message, error.hint)

        with self.subTest('Test setting dotted path is not subclass of ModelAdmin'):
            path = 'openwisp_notifications.admin.NotificationSettingInline'
            with patch.object(app_settings, 'IGNORE_ENABLED_ADMIN', [path]):
                error_message = (
                    f'"{path}" does not subclasses "django.contrib.admin.ModelAdmin"'
                )
                error = run_check()
                self.assertIn(error_message, error.hint)



# Generated by Django 3.0.8 on 2020-08-18 13:09

import uuid

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('openwisp_notifications', '0005_delete_notificationuser'),
    ]

    operations = [
        migrations.CreateModel(
            name='IgnoreObjectNotification',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ('object_id', models.CharField(max_length=255)),
                ('valid_till', models.DateTimeField(null=True)),
                (
                    'object_content_type',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to='contenttypes.ContentType',
                    ),
                ),
                (
                    'user',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                'abstract': False,
                'swappable': 'OPENWISP_NOTIFICATIONS_OBJECTNOTIFICATION_MODEL',
                'ordering': ['valid_till'],
            },
        ),
    ]



import logging
from contextlib import contextmanager

from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.contrib.sites.models import Site
from django.core.cache import cache
from django.db import models
from django.db.models.constraints import UniqueConstraint
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.html import mark_safe
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _
from markdown import markdown
from notifications.base.models import AbstractNotification as BaseNotification
from swapper import get_model_name

from openwisp_notifications import settings as app_settings
from openwisp_notifications.exceptions import NotificationRenderException
from openwisp_notifications.types import (
    NOTIFICATION_CHOICES,
    get_notification_configuration,
)
from openwisp_notifications.utils import _get_absolute_url, _get_object_link
from openwisp_utils.base import UUIDModel

logger = logging.getLogger(__name__)


@contextmanager
def notification_render_attributes(obj, **attrs):
    """
    This context manager sets temporary attributes on
    the notification object to allowing rendering of
    notification.

    It can only be used to set aliases of the existing attributes.
    By default, it will set the following aliases:
        - actor_link -> actor_url
        - action_link -> action_url
        - target_link -> target_url
    """
    defaults = {
        'actor_link': 'actor_url',
        'action_link': 'action_url',
        'target_link': 'target_url',
    }
    defaults.update(attrs)

    for target_attr, source_attr in defaults.items():
        setattr(obj, target_attr, getattr(obj, source_attr))

    yield obj

    for attr in defaults.keys():
        delattr(obj, attr)


class AbstractNotification(UUIDModel, BaseNotification):
    CACHE_KEY_PREFIX = 'ow-notifications-'
    type = models.CharField(max_length=30, null=True, choices=NOTIFICATION_CHOICES)
    _actor = BaseNotification.actor
    _action_object = BaseNotification.action_object
    _target = BaseNotification.target

    class Meta(BaseNotification.Meta):
        abstract = True

    def __init__(self, *args, **kwargs):
        related_objs = [
            (opt, kwargs.pop(opt, None)) for opt in ('target', 'action_object', 'actor')
        ]
        super().__init__(*args, **kwargs)
        for opt, obj in related_objs:
            if obj is not None:
                setattr(self, f'{opt}_object_id', obj.pk)
                setattr(
                    self,
                    f'{opt}_content_type',
                    ContentType.objects.get_for_model(obj),
                )

    def __str__(self):
        return self.timesince()

    @classmethod
    def _cache_key(cls, *args):
        args = map(str, args)
        key = '-'.join(args)
        return f'{cls.CACHE_KEY_PREFIX}{key}'

    @classmethod
    def count_cache_key(cls, user_pk):
        return cls._cache_key(f'unread-{user_pk}')

    @classmethod
    def invalidate_unread_cache(cls, user):
        """
        Invalidate unread cache for user.
        """
        cache.delete(cls.count_cache_key(user.pk))

    def _get_related_object_url(self, field):
        """
        Returns URLs for "actor", "action_object" and "target" fields.
        """
        if self.type:
            # Generate URL according to the notification configuration
            config = get_notification_configuration(self.type)
            url = config.get(f'{field}_link', None)
            if url:
                try:
                    url_callable = import_string(url)
                    return url_callable(self, field=field, absolute_url=True)
                except ImportError:
                    return url
        return _get_object_link(self, field=field, absolute_url=True)

    @property
    def actor_url(self):
        return self._get_related_object_url(field='actor')

    @property
    def action_url(self):
        return self._get_related_object_url(field='action_object')

    @property
    def target_url(self):
        return self._get_related_object_url(field='target')

    @cached_property
    def message(self):
        with notification_render_attributes(self):
            return self.get_message()

    @cached_property
    def rendered_description(self):
        if not self.description:
            return
        with notification_render_attributes(self):
            data = self.data or {}
            desc = self.description.format(notification=self, **data)
        return mark_safe(markdown(desc))

    @property
    def email_message(self):
        with notification_render_attributes(self, target_link='redirect_view_url'):
            return self.get_message()

    def get_message(self):
        if not self.type:
            return self.description
        try:
            config = get_notification_configuration(self.type)
            data = self.data or {}
            if 'message' in data:
                md_text = data['message'].format(notification=self, **data)
            elif 'message' in config:
                md_text = config['message'].format(notification=self, **data)
            else:
                md_text = render_to_string(
                    config['message_template'], context=dict(notification=self, **data)
                ).strip()
        except (AttributeError, KeyError, NotificationRenderException) as exception:
            self._invalid_notification(
                self.pk,
                exception,
                'Error encountered in rendering notification message',
            )
        return mark_safe(markdown(md_text))

    @cached_property
    def email_subject(self):
        if self.type:
            try:
                config = get_notification_configuration(self.type)
                data = self.data or {}
                return config['email_subject'].format(
                    site=Site.objects.get_current(), notification=self, **data
                )
            except (AttributeError, KeyError, NotificationRenderException) as exception:
                self._invalid_notification(
                    self.pk,
                    exception,
                    'Error encountered in generating notification email',
                )
        elif self.data.get('email_subject', None):
            return self.data.get('email_subject')
        else:
            return self.message

    def _related_object(self, field):
        obj_id = getattr(self, f'{field}_object_id')
        obj_content_type_id = getattr(self, f'{field}_content_type_id')
        if not obj_id:
            return
        cache_key = self._cache_key(obj_content_type_id, obj_id)
        obj = cache.get(cache_key)
        if not obj:
            obj = getattr(self, f'_{field}')
            cache.set(
                cache_key,
                obj,
                timeout=app_settings.CACHE_TIMEOUT,
            )
        return obj

    def _invalid_notification(self, pk, exception, error_message):
        from openwisp_notifications.tasks import delete_notification

        logger.error(exception)
        delete_notification.delay(notification_id=pk)
        if isinstance(exception, NotificationRenderException):
            raise exception
        raise NotificationRenderException(error_message)

    @cached_property
    def actor(self):
        return self._related_object('actor')

    @cached_property
    def action_object(self):
        return self._related_object('action_object')

    @cached_property
    def target(self):
        return self._related_object('target')

    @property
    def redirect_view_url(self):
        return _get_absolute_url(
            reverse('notifications:notification_read_redirect', args=(self.pk,))
        )


class AbstractNotificationSetting(UUIDModel):
    _RECEIVE_HELP = (
        'Note: Non-superadmin users receive '
        'notifications only for organizations '
        'of which they are member of.'
    )
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    type = models.CharField(
        max_length=30,
        null=True,
        choices=NOTIFICATION_CHOICES,
        verbose_name='Notification Type',
    )
    organization = models.ForeignKey(
        get_model_name('openwisp_users', 'Organization'),
        on_delete=models.CASCADE,
    )
    web = models.BooleanField(
        _('web notifications'), null=True, blank=True, help_text=_(_RECEIVE_HELP)
    )
    email = models.BooleanField(
        _('email notifications'), null=True, blank=True, help_text=_(_RECEIVE_HELP)
    )
    deleted = models.BooleanField(_('Delete'), null=True, blank=True, default=False)

    class Meta:
        abstract = True
        constraints = [
            UniqueConstraint(
                fields=['organization', 'type', 'user'],
                name='unique_notification_setting',
            ),
        ]
        verbose_name = _('user notification settings')
        verbose_name_plural = verbose_name
        ordering = ['organization', 'type']
        indexes = [
            models.Index(fields=['type', 'organization']),
        ]

    def __str__(self):
        return '{type} - {organization}'.format(
            type=self.type_config['verbose_name'],
            organization=self.organization,
        )

    def save(self, *args, **kwargs):
        if not self.web_notification:
            self.email = self.web_notification
        return super().save(*args, **kwargs)

    def full_clean(self, *args, **kwargs):
        if self.email == self.type_config['email_notification']:
            self.email = None
        if self.web == self.type_config['web_notification']:
            self.web = None
        return super().full_clean(*args, **kwargs)

    @property
    def type_config(self):
        return get_notification_configuration(self.type)

    @property
    def email_notification(self):
        if self.email is not None:
            return self.email
        return self.type_config.get('email_notification')

    @property
    def web_notification(self):
        if self.web is not None:
            return self.web
        return self.type_config.get('web_notification')


class AbstractIgnoreObjectNotification(UUIDModel):
    """
    This model stores information about ignoring notification
    from a specific object for a user. Any instance of the model
    should be only stored until "valid_till" expires.
    """

    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    object_content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.CharField(max_length=255)
    object = GenericForeignKey('object_content_type', 'object_id')
    valid_till = models.DateTimeField(null=True)

    class Meta:
        abstract = True
        ordering = ['valid_till']



from datetime import timedelta

from celery import shared_task
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from django.db.utils import OperationalError
from django.utils import timezone

from openwisp_notifications import types
from openwisp_notifications.swapper import load_model, swapper_load_model
from openwisp_utils.tasks import OpenwispCeleryTask

User = get_user_model()

Notification = load_model('Notification')
NotificationSetting = load_model('NotificationSetting')
IgnoreObjectNotification = load_model('IgnoreObjectNotification')

Organization = swapper_load_model('openwisp_users', 'Organization')
OrganizationUser = swapper_load_model('openwisp_users', 'OrganizationUser')


@shared_task(base=OpenwispCeleryTask)
def delete_obsolete_objects(instance_app_label, instance_model, instance_id):
    """
    Delete Notification and IgnoreObjectNotification objects having
    instance' as related objects..
    """
    try:
        instance_content_type = ContentType.objects.get_by_natural_key(
            instance_app_label, instance_model
        )
    except ContentType.DoesNotExist:
        return
    else:
        # Delete Notification objects
        where = (
            Q(actor_content_type=instance_content_type)
            | Q(action_object_content_type=instance_content_type)
            | Q(target_content_type=instance_content_type)
        )
        where = where & (
            Q(actor_object_id=instance_id)
            | Q(action_object_object_id=instance_id)
            | Q(target_object_id=instance_id)
        )
        Notification.objects.filter(where).delete()

        # Delete IgnoreObjectNotification objects
        try:
            IgnoreObjectNotification.objects.filter(
                object_id=instance_id, object_content_type_id=instance_content_type.pk
            ).delete()
        except OperationalError:
            # Raised when an object is deleted in migration
            return


@shared_task(base=OpenwispCeleryTask)
def delete_notification(notification_id):
    Notification.objects.filter(pk=notification_id).delete()


@shared_task
def delete_old_notifications(days):
    """
    Delete notifications having 'timestamp' more than "days" days.
    """
    Notification.objects.filter(
        timestamp__lte=timezone.now() - timedelta(days=days)
    ).delete()


# Following tasks updates notification settings in database.
# 'ns' is short for notification_setting
def create_notification_settings(user, organizations, notification_types):
    for type in notification_types:
        for org in organizations:
            NotificationSetting.objects.update_or_create(
                defaults={'deleted': False}, user=user, type=type, organization=org
            )


@shared_task(base=OpenwispCeleryTask)
def update_superuser_notification_settings(instance_id, is_superuser, is_created):
    """
    Adds notification setting for all notification types and organizations.
    If a superuser gets demoted, flags it's notification settings as deleted.
    """
    user = User.objects.get(pk=instance_id)

    # When a user is demoted from superuser status,
    # only keep notification settings for organization they are member of.
    if not (is_superuser or is_created):
        NotificationSetting.objects.filter(user_id=instance_id).exclude(
            organization__in=user.organizations_managed
        ).update(deleted=True)
        return

    if not is_superuser:
        return

    # Create notification settings for superuser
    create_notification_settings(
        user=user,
        organizations=Organization.objects.all(),
        notification_types=types.NOTIFICATION_TYPES.keys(),
    )


@shared_task(base=OpenwispCeleryTask)
def ns_register_unregister_notification_type(
    notification_type=None, delete_unregistered=True
):
    """
    Creates notification setting for registered notification types.
    Deletes notification for unregistered notification types.
    """

    notification_types = (
        [notification_type] if notification_type else types.NOTIFICATION_TYPES.keys()
    )

    organizations = Organization.objects.all()
    # Create notification settings for superusers
    for user in User.objects.filter(is_superuser=True).iterator():
        create_notification_settings(user, organizations, notification_types)

    # Create notification settings for organization admin
    for org_user in OrganizationUser.objects.select_related(
        'user', 'organization'
    ).filter(is_admin=True, user__is_superuser=False):
        create_notification_settings(
            org_user.user, [org_user.organization], notification_types
        )

    if delete_unregistered:
        # Delete all notification settings for unregistered notification types
        NotificationSetting.objects.exclude(type__in=notification_types).update(
            deleted=True
        )
        # Delete notifications related to unregister notification types
        Notification.objects.exclude(type__in=notification_types).delete()


@shared_task(base=OpenwispCeleryTask)
def update_org_user_notificationsetting(org_user_id, user_id, org_id, is_org_admin):
    """
    Adds notification settings for all notification types when a new
    organization user is added.
    """
    user = User.objects.get(pk=user_id)
    if not user.is_superuser:
        # The following query covers conditions for change in admin status
        # and organization field of related OrganizationUser objects
        NotificationSetting.objects.filter(user=user).exclude(
            organization_id__in=user.organizations_managed
        ).update(deleted=True)

    if not is_org_admin:
        return

    # Create new notification settings
    organization = Organization.objects.get(id=org_id)
    create_notification_settings(
        user=user,
        organizations=[organization],
        notification_types=types.NOTIFICATION_TYPES.keys(),
    )


@shared_task(base=OpenwispCeleryTask)
def ns_organization_user_deleted(user_id, org_id):
    """
    Deletes notification settings for all notification types when
    an organization user is deleted.
    """
    NotificationSetting.objects.filter(user_id=user_id, organization_id=org_id).update(
        deleted=True
    )


@shared_task(base=OpenwispCeleryTask)
def ns_organization_created(instance_id):
    """
    Adds notification setting of all registered types
    for a newly created organization.
    """
    organization = Organization.objects.get(id=instance_id)
    for user in User.objects.filter(is_superuser=True):
        create_notification_settings(
            user=user,
            organizations=[organization],
            notification_types=types.NOTIFICATION_TYPES.keys(),
        )


@shared_task(base=OpenwispCeleryTask)
def delete_ignore_object_notification(instance_id):
    """
    Deletes IgnoreObjectNotification object post it's expiration.
    """
    IgnoreObjectNotification.objects.filter(id=instance_id).delete()



# Generated by Django 3.0.8 on 2020-07-24 23:56

from django.contrib.auth.management import create_permissions
from django.db import migrations


def delete_notification_user_permissions(apps, schema_editor):
    Permission = apps.get_model('auth', 'Permission')
    Permission.objects.filter(
        content_type__app_label='openwisp_notifications',
        content_type__model='notificationuser',
    ).delete()


def populate_notification_user_permissions(apps, schema_editor):
    # Populate Permissions
    app_config = apps.get_app_config('openwisp_notifications')
    app_config.models_module = True
    create_permissions(app_config, apps=apps, verbosity=0)
    app_config.models_module = None


class Migration(migrations.Migration):
    dependencies = [
        ('openwisp_notifications', '0004_notificationsetting'),
    ]

    operations = [
        migrations.RunPython(
            delete_notification_user_permissions,
            reverse_code=populate_notification_user_permissions,
        ),
        migrations.DeleteModel(name='NotificationUser'),
    ]



import uuid
from unittest.mock import patch

from django.contrib.admin.sites import AdminSite
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission
from django.core.cache import cache
from django.forms.widgets import MediaOrderConflictWarning
from django.test import TestCase, override_settings, tag
from django.urls import reverse

from openwisp_notifications import settings as app_settings
from openwisp_notifications.admin import NotificationSettingInline
from openwisp_notifications.signals import notify
from openwisp_notifications.swapper import load_model, swapper_load_model
from openwisp_notifications.widgets import _add_object_notification_widget
from openwisp_users.admin import UserAdmin
from openwisp_users.tests.utils import TestMultitenantAdminMixin

from .test_helpers import MessagingRequest

Notification = load_model('Notification')
NotificationSetting = load_model('NotificationSetting')
notification_queryset = Notification.objects.order_by('-timestamp')
Group = swapper_load_model('openwisp_users', 'Group')


class MockUser:
    def __init__(self, is_superuser=False):
        self.is_superuser = is_superuser
        self.id = uuid.uuid4()

    def has_perm(self, perm):
        return True

    @property
    def pk(self):
        return self.id


User = get_user_model()
su_request = MessagingRequest()
su_request.user = MockUser(is_superuser=True)

op_request = MessagingRequest()
op_request.user = MockUser(is_superuser=False)


class BaseTestAdmin(TestMultitenantAdminMixin, TestCase):
    def _login_admin(self):
        u = User.objects.create_superuser('admin', 'admin', 'test@test.com')
        self.client.force_login(u)
        return u

    def setUp(self):
        self.admin = self._login_admin()
        self.notification_options = dict(
            sender=self.admin,
            recipient=self.admin,
            description='Test Notification',
            verb='Test Notification',
            email_subject='Test Email subject',
            url='localhost:8000/admin',
        )
        self.site = AdminSite()
        self.ns_inline = NotificationSettingInline(NotificationSetting, self.site)

    @property
    def _url(self):
        return reverse('admin:index')

    def _expected_output(self, count=None):
        if count:
            return '<span id="ow-notification-count">{0}</span>'.format(count)
        return 'id="openwisp_notifications">'


class TestAdmin(BaseTestAdmin):
    """
    Tests notifications in admin
    """

    app_label = 'openwisp_notifications'

    def test_zero_notifications(self):
        r = self.client.get(self._url)
        self.assertContains(r, self._expected_output())

    def test_non_zero_notifications(self):
        patched_function = 'openwisp_notifications.templatetags.notification_tags._get_user_unread_count'
        with self.subTest("Test UI for less than 100 notifications"):
            with patch(patched_function, return_value=10):
                r = self.client.get(self._url)
                self.assertContains(r, self._expected_output('10'))

        Notification.invalidate_unread_cache(self.admin)

        with self.subTest("Test UI for 99+ notifications"):
            with patch(patched_function, return_value=100):
                r = self.client.get(self._url)
                self.assertContains(r, self._expected_output('99+'))

    def test_cached_value(self):
        self.client.get(self._url)
        cache_key = Notification.count_cache_key(self.admin.pk)
        self.assertEqual(cache.get(cache_key), 0)
        return cache_key

    def test_cached_invalidation(self):
        cache_key = self.test_cached_value()
        notify.send(**self.notification_options)
        self.assertIsNone(cache.get(cache_key))
        self.client.get(self._url)
        self.assertEqual(cache.get(cache_key), 1)

    @tag('skip_prod')
    # This tests depends on the static storage backend of the project.
    # In prod environment, the filenames could get changed due to
    # static minification and cache invalidation. Hence, these tests
    # should not be run on prod environment because they'll fail.
    def test_default_notification_setting(self):
        res = self.client.get(self._url)
        self.assertContains(
            res, '/static/openwisp-notifications/audio/notification_bell.mp3'
        )
        self.assertContains(res, 'window.location')

    @tag('skip_prod')
    # For more info, look at TestAdmin.test_default_notification_setting
    @patch.object(
        app_settings,
        'SOUND',
        '/static/notification.mp3',
    )
    def test_notification_sound_setting(self):
        res = self.client.get(self._url)
        self.assertContains(res, '/static/notification.mp3')
        self.assertNotContains(
            res, '/static/openwisp-notifications/audio/notification_bell.mp3'
        )

    @patch.object(
        app_settings,
        'HOST',
        'https://example.com',
    )
    def test_notification_host_setting(self):
        res = self.client.get(self._url)
        self.assertContains(res, 'https://example.com')
        self.assertNotContains(res, 'window.location')

    def test_login_load_javascript(self):
        self.client.logout()
        response = self.client.get(reverse('admin:login'))
        self.assertNotContains(response, 'notifications.js')

    def test_websocket_protocol(self):
        with self.subTest('Test in production environment'):
            response = self.client.get(self._url)
            self.assertContains(response, 'wss')

    def test_notification_setting_inline_read_only_fields(self):
        with self.subTest('Test for superuser'):
            self.assertListEqual(self.ns_inline.get_readonly_fields(su_request), [])

        with self.subTest('Test for non-superuser'):
            self.assertListEqual(
                self.ns_inline.get_readonly_fields(op_request),
                ['type', 'organization'],
            )

    def test_notification_setting_inline_add_permission(self):
        with self.subTest('Test for superuser'):
            self.assertTrue(self.ns_inline.has_add_permission(su_request))

        with self.subTest('Test for non-superuser'):
            self.assertFalse(
                self.ns_inline.has_add_permission(op_request),
            )

    def test_notification_setting_inline_delete_permission(self):
        with self.subTest('Test for superuser'):
            self.assertTrue(self.ns_inline.has_delete_permission(su_request))

        with self.subTest('Test for non-superuser'):
            self.assertFalse(self.ns_inline.has_delete_permission(op_request))

    def test_notification_setting_inline_organization_formfield(self):
        response = self.client.get(
            reverse('admin:openwisp_users_user_change', args=(self.admin.pk,))
        )
        organization = self._get_org(org_name='default')
        self.assertContains(
            response,
            f'<option value="{organization.id}">{organization.name}</option>',
        )

    def test_notification_setting_inline_admin_has_change_permission(self):
        with self.subTest('Test for superuser'):
            self.assertTrue(
                self.ns_inline.has_change_permission(su_request),
            )

        with self.subTest('Test for non-superuser'):
            self.assertFalse(
                self.ns_inline.has_change_permission(op_request),
            )
            self.assertTrue(
                self.ns_inline.has_change_permission(op_request, obj=op_request.user),
            )

    def test_org_admin_view_same_org_user_notification_setting(self):
        org_owner = self._create_org_user(
            user=self._get_operator(),
            is_admin=True,
        )
        org_admin = self._create_org_user(
            user=self._create_user(
                username='user', email='user@user.com', is_staff=True
            ),
            is_admin=True,
        )
        permissions = Permission.objects.all()
        org_owner.user.user_permissions.set(permissions)
        org_admin.user.user_permissions.set(permissions)
        self.client.force_login(org_owner.user)

        response = self.client.get(
            reverse('admin:openwisp_users_user_change', args=(org_admin.user_id,)),
        )
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'User notification settings')
        self.assertNotContains(
            response, '<option value="default" selected>Default Type</option>'
        )

    def test_ignore_notification_widget_add_view(self):
        url = reverse('admin:openwisp_users_organization_add')
        response = self.client.get(url)
        self.assertNotContains(response, 'owIsChangeForm')


@tag('skip_prod')
# For more info, look at TestAdmin.test_default_notification_setting
class TestAdminMedia(BaseTestAdmin):
    """
    Tests notifications admin media
    """

    def test_jquery_import(self):
        response = self.client.get(self._url)
        self.assertInHTML(
            '<script src="/static/admin/js/jquery.init.js">',
            str(response.content),
            1,
        )
        self.assertInHTML(
            '<script src="/static/admin/js/vendor/jquery/jquery.min.js">',
            str(response.content),
            1,
        )

        response = self.client.get(reverse('admin:sites_site_changelist'))
        self.assertIn(
            '/static/admin/js/jquery.init.js',
            str(response.content),
            1,
        )
        self.assertIn(
            '/static/admin/js/vendor/jquery/jquery.min.js',
            str(response.content),
            1,
        )

    def test_object_notification_setting_empty(self):
        response = self.client.get(
            reverse('admin:openwisp_users_user_change', args=(self.admin.pk,))
        )
        self.assertNotContains(
            response, 'src="/static/openwisp-notifications/js/object-notifications.js"'
        )

    @override_settings(
        OPENWISP_NOTIFICATIONS_IGNORE_ENABLED_ADMIN=['openwisp_users.admin.UserAdmin'],
    )
    def test_object_notification_setting_configured(self):
        _add_object_notification_widget()
        response = self.client.get(
            reverse('admin:openwisp_users_user_change', args=(self.admin.pk,))
        )
        self.assertContains(
            response,
            'src="/static/openwisp-notifications/js/object-notifications.js"',
            1,
        )

        # If a ModelAdmin already has a Media class
        with self.assertWarns(MediaOrderConflictWarning):
            _add_object_notification_widget()
            response = self.client.get(
                reverse('admin:openwisp_users_user_change', args=(self.admin.pk,))
            )

        # If a ModelAdmin has list instances of js and css
        UserAdmin.Media.css = {'all': list()}
        UserAdmin.Media.js = list()
        _add_object_notification_widget()
        response = self.client.get(
            reverse('admin:openwisp_users_user_change', args=(self.admin.pk,))
        )

        # If ModelAdmin has empty attributes
        UserAdmin.Media.js = []
        UserAdmin.Media.css = {}
        _add_object_notification_widget()
        response = self.client.get(
            reverse('admin:openwisp_users_user_change', args=(self.admin.pk,))
        )
        UserAdmin.Media = None



from django.conf import settings
from django.utils.module_loading import import_string

from openwisp_utils.admin import UUIDAdmin


class IgnoreObjectNotificationWidgetMedia:
    extend = True
    js = ('openwisp-notifications/js/object-notifications.js',)
    css = {'all': ('openwisp-notifications/css/object-notifications.css',)}


def _add_object_notification_widget():
    """
    Adds object notification widget on configured ModelAdmins.
    """
    IGNORE_ENABLED_ADMIN = getattr(
        settings, 'OPENWISP_NOTIFICATIONS_IGNORE_ENABLED_ADMIN', []
    )
    for model_admin_path in IGNORE_ENABLED_ADMIN:
        model_admin_class = import_string(model_admin_path)
        try:
            if isinstance(model_admin_class.Media.js, list):
                model_admin_class.Media.js.extend(
                    IgnoreObjectNotificationWidgetMedia.js
                )
            elif isinstance(model_admin_class.Media.js, tuple):
                model_admin_class.Media.js += IgnoreObjectNotificationWidgetMedia.js

            if 'all' in model_admin_class.Media.css:
                if isinstance(model_admin_class.Media.css['all'], list):
                    model_admin_class.Media.css['all'].extend(
                        IgnoreObjectNotificationWidgetMedia.css['all']
                    )
                elif isinstance(model_admin_class.Media.css['all'], tuple):
                    model_admin_class.Media.css[
                        'all'
                    ] += IgnoreObjectNotificationWidgetMedia.css['all']
            else:
                model_admin_class.Media.css.update(
                    IgnoreObjectNotificationWidgetMedia.css
                )
        except AttributeError:
            model_admin_class.Media = IgnoreObjectNotificationWidgetMedia
            # Needed tp maintain order or JS imports.
            model_admin_class.Media.js = UUIDAdmin.Media.js + model_admin_class.Media.js



from django.urls import path

from openwisp_notifications.api import views

app_name = 'openwisp_notifications'


def get_api_urls(api_views=None):
    if not api_views:
        api_views = views
    return [
        path('', views.notifications_list, name='notifications_list'),
        path('read/', views.notifications_read_all, name='notifications_read_all'),
        path('<uuid:pk>/', views.notification_detail, name='notification_detail'),
        path(
            '<uuid:pk>/redirect/',
            views.notification_read_redirect,
            name='notification_read_redirect',
        ),
        path(
            'user-setting/',
            views.notification_setting_list,
            name='notification_setting_list',
        ),
        path(
            'user-setting/<uuid:pk>/',
            views.notification_setting,
            name='notification_setting',
        ),
        path(
            'ignore/',
            views.ignore_object_notification_list,
            name='ignore_object_notification_list',
        ),
        path(
            'ignore/<str:app_label>/<str:model_name>/<uuid:object_id>/',
            views.ignore_object_notification,
            name='ignore_object_notification',
        ),
    ]



from django.urls import include, path

from .api.urls import get_api_urls


def get_urls(api_views=None, social_views=None):
    """
    Returns a list of urlpatterns
    Arguments:
        api_views(optional): views for Notifications API
    """
    urls = [
        path('api/v1/notifications/notification/', include(get_api_urls(api_views)))
    ]
    return urls


app_name = 'notifications'
urlpatterns = get_urls()



from swapper import swappable_setting

from openwisp_notifications.base.models import (
    AbstractIgnoreObjectNotification,
    AbstractNotification,
    AbstractNotificationSetting,
)


class Notification(AbstractNotification):
    class Meta(AbstractNotification.Meta):
        abstract = False
        app_label = 'openwisp_notifications'
        swappable = swappable_setting('openwisp_notifications', 'Notification')


class NotificationSetting(AbstractNotificationSetting):
    class Meta(AbstractNotificationSetting.Meta):
        abstract = False
        swappable = swappable_setting('openwisp_notifications', 'NotificationSetting')


class IgnoreObjectNotification(AbstractIgnoreObjectNotification):
    class Meta(AbstractIgnoreObjectNotification.Meta):
        abstract = False
        swappable = swappable_setting(
            'openwisp_notifications', 'IgnoreObjectNotification'
        )



import swapper


def get_swapped_model(apps, app_name, model_name):
    model_path = swapper.get_model_name(app_name, model_name)
    app, model = swapper.split(model_path)
    return apps.get_model(app, model)



from django.apps import AppConfig
from django.db.models.signals import post_migrate
from django.utils.translation import gettext_lazy as _

from . import settings as app_settings


class OpenwispNotificationsConfig(AppConfig):
    name = 'openwisp_notifications'
    label = 'openwisp_notifications'
    verbose_name = _('Notifications')

    def ready(self):
        from openwisp_notifications.handlers import (
            notification_type_registered_unregistered_handler,
            notify_handler,
        )
        from openwisp_notifications.signals import notify

        notify.connect(
            notify_handler, dispatch_uid='openwisp_notifications.model.notifications'
        )
        if app_settings.POPULATE_PREFERENCES_ON_MIGRATE:
            post_migrate.connect(
                notification_type_registered_unregistered_handler,
                sender=self,
                dispatch_uid='register_unregister_notification_types',
            )

        # Add CORS configuration checks
        from openwisp_notifications.checks import check_cors_configuration  # noqa



from asgiref.sync import async_to_sync
from channels import layers
from django.core.cache import cache
from django.utils.timezone import now, timedelta

from openwisp_notifications.api.serializers import NotFound, NotificationListSerializer
from openwisp_notifications.utils import normalize_unread_count

from .. import settings as app_settings
from ..swapper import load_model

Notification = load_model('Notification')


def user_in_notification_storm(user):
    """
    A user is affected by notifications storm if any of short term
    or long term check passes. The checks are configured by
    "OPENWISP_NOTIFICATIONS_NOTIFICATION_STORM_PREVENTION" setting.
    If the user is found to be affected by a notification storm,
    the value of this function is cached for 60 seconds.
    """
    in_notification_storm = cache.get(f'ow-noti-storm-{user.pk}', False)
    if in_notification_storm:
        return True
    if (
        user.notifications.filter(
            timestamp__gte=now()
            - timedelta(
                seconds=app_settings.NOTIFICATION_STORM_PREVENTION[
                    'short_term_time_period'
                ]
            )
        ).count()
        > app_settings.NOTIFICATION_STORM_PREVENTION['short_term_notification_count']
    ):
        in_notification_storm = True
    elif (
        user.notifications.filter(
            timestamp__gte=now()
            - timedelta(
                seconds=app_settings.NOTIFICATION_STORM_PREVENTION[
                    'long_term_time_period'
                ]
            )
        ).count()
        > app_settings.NOTIFICATION_STORM_PREVENTION['long_term_notification_count']
    ):
        in_notification_storm = True
    if in_notification_storm:
        cache.set(f'ow-noti-storm-{user.pk}', True, 60)
    return in_notification_storm


def notification_update_handler(reload_widget=False, notification=None, recipient=None):
    channel_layer = layers.get_channel_layer()
    try:
        assert notification is not None
        notification = NotificationListSerializer(notification).data
    except (NotFound, AssertionError):
        pass
    async_to_sync(channel_layer.group_send)(
        f'ow-notification-{recipient.pk}',
        {
            'type': 'send.updates',
            'reload_widget': reload_widget,
            'notification': notification,
            'recipient': str(recipient.pk),
            'in_notification_storm': user_in_notification_storm(recipient),
            'notification_count': normalize_unread_count(
                recipient.notifications.unread().count()
            ),
        },
    )



from django.db.models.signals import post_save
from django.test import TransactionTestCase

from openwisp_notifications.handlers import (
    notification_type_registered_unregistered_handler,
)
from openwisp_notifications.swapper import load_model, swapper_load_model
from openwisp_notifications.tests.test_helpers import (
    base_register_notification_type,
    base_unregister_notification_type,
    mock_notification_types,
    register_notification_type,
)
from openwisp_users.tests.utils import TestOrganizationMixin

test_notification_type = {
    'verbose_name': 'Test Notification Type',
    'level': 'test',
    'verb': 'testing',
    'message': '{notification.verb} initiated by {notification.actor} since {notification}',
    'email_subject': '[{site.name}] {notification.verb} reported by {notification.actor}',
}

NotificationSetting = load_model('NotificationSetting')
Organization = swapper_load_model('openwisp_users', 'Organization')
OrganizationUser = swapper_load_model('openwisp_users', 'OrganizationUser')

ns_queryset = NotificationSetting.objects.filter(type='default')


class TestNotificationSetting(TestOrganizationMixin, TransactionTestCase):
    def setUp(self):
        self.default_org = self._get_org('default')

    def _create_staff_org_admin(self):
        return self._create_org_user(user=self._create_operator(), is_admin=True)

    def test_no_user(self):
        self.assertEqual(ns_queryset.count(), 0)

    def test_superuser_created(self):
        admin = self._get_admin()
        self.assertEqual(ns_queryset.filter(user=admin).count(), 1)

    def test_user_created(self):
        self._get_user()
        self.assertEqual(ns_queryset.count(), 0)

    @mock_notification_types
    def test_notification_type_registered(self):
        register_notification_type('test', test_notification_type)
        queryset = NotificationSetting.objects.filter(type='test')

        self._get_user()
        self.assertEqual(queryset.count(), 0)

        self._get_admin()
        self.assertEqual(queryset.count(), 1)
        self.assertEquals(
            queryset.first().__str__(), 'Test Notification Type - default'
        )

    def test_organization_created_no_initial_user(self):
        org = self._get_org()
        queryset = ns_queryset.filter(organization=org)
        self.assertEqual(ns_queryset.count(), 0)

        # Notification setting is not created for normal user
        self._get_user()
        self.assertEqual(queryset.count(), 0)

        self._get_admin()
        self.assertEqual(queryset.count(), 1)

    def test_organization_user(self):
        karen = self._get_user()
        ken = self._create_user(username='ken', email='ken@ken.com')
        org = self._get_org()
        OrganizationUser.objects.create(user=karen, organization=org, is_admin=True)
        org_user = OrganizationUser.objects.create(
            user=ken, organization=org, is_admin=True
        )
        self.assertEqual(ns_queryset.count(), 2)
        org_user.delete()
        # OrganizationOwner can not be deleted before transferring ownership
        self.assertEqual(ns_queryset.filter(deleted=False).count(), 1)
        self.assertEqual(ns_queryset.filter(deleted=True).count(), 1)

    @mock_notification_types
    def test_register_notification_org_user(self):
        self._create_staff_org_admin()

        queryset = NotificationSetting.objects.filter(type='test')
        self.assertEqual(queryset.count(), 0)
        register_notification_type('test', test_notification_type)
        self.assertEqual(queryset.count(), 1)

    @mock_notification_types
    def test_post_migration_handler(self):
        from openwisp_notifications.types import NOTIFICATION_CHOICES

        # Simulates loading of app when Django server starts
        admin = self._get_admin()
        org_user = self._create_staff_org_admin()
        self.assertEqual(ns_queryset.count(), 3)

        base_unregister_notification_type('default')
        base_register_notification_type('test', test_notification_type)
        notification_type_registered_unregistered_handler(sender=self)

        # Notification Setting for "default" type are deleted
        self.assertEqual(ns_queryset.filter(type='default', deleted=True).count(), 3)

        # Notification Settings for "test" type are created
        queryset = NotificationSetting.objects.filter(deleted=False)
        notification_types_count = len(NOTIFICATION_CHOICES)
        self.assertEqual(queryset.count(), 3 * notification_types_count)
        self.assertEqual(
            queryset.filter(user=admin).count(), 2 * notification_types_count
        )
        self.assertEqual(
            queryset.filter(user=org_user.user).count(), 1 * notification_types_count
        )

    def test_superuser_demoted_to_user(self):
        admin = self._get_admin()
        admin.is_superuser = False
        admin.save()

        self.assertEqual(ns_queryset.filter(deleted=True).count(), 1)

    def test_user_promoted_to_superuser(self):
        user = self._create_operator()
        self.assertEqual(ns_queryset.count(), 0)

        user.is_superuser = True
        user.save()

        self.assertEqual(ns_queryset.count(), 1)

    def test_superuser_demoted_to_org_admin(self):
        admin = self._get_admin()
        admin.is_superuser = False
        admin.save()
        org = Organization.objects.get(name='default')
        OrganizationUser.objects.create(user=admin, organization=org, is_admin=True)

        self.assertEqual(ns_queryset.count(), 1)

    def test_org_admin_demoted_to_org_user(self):
        org_user = self._create_staff_org_admin()
        self.assertEqual(ns_queryset.count(), 1)
        org_user.organizationowner.delete()
        org_user.is_admin = False
        org_user.full_clean()
        org_user.save()
        self.assertEqual(ns_queryset.filter(deleted=False).count(), 0)
        self.assertEqual(ns_queryset.filter(deleted=True).count(), 1)

    def test_org_user_promoted_to_org_admin(self):
        org_user = self._create_org_user(user=self._create_operator(), is_admin=False)
        self.assertEqual(ns_queryset.count(), 0)
        org_user.is_admin = True
        org_user.full_clean()
        org_user.save()
        self.assertEqual(ns_queryset.count(), 1)

    def test_multiple_org_membership(self):
        user = self._get_user()
        default_org = Organization.objects.first()
        test_org = self._get_org()
        self.assertEqual(ns_queryset.count(), 0)

        OrganizationUser.objects.create(
            user=user, organization=default_org, is_admin=True
        )
        self.assertEqual(ns_queryset.count(), 1)

        OrganizationUser.objects.create(user=user, organization=test_org, is_admin=True)
        self.assertEqual(ns_queryset.count(), 2)

    @mock_notification_types
    def test_notification_setting_full_clean(self):
        test_type = {
            'verbose_name': 'Test Notification Type',
            'level': 'info',
            'verb': 'testing',
            'message': 'Test message',
            'email_subject': 'Test Email Subject',
            'web_notification': False,
            'email_notification': False,
        }
        register_notification_type('test_type', test_type)
        self._get_admin()
        queryset = NotificationSetting.objects.filter(type='test_type')
        queryset.update(email=False, web=False)
        notification_setting = queryset.first()

        notification_setting.full_clean()
        self.assertIsNone(notification_setting.email)
        self.assertIsNone(notification_setting.web)

    def test_organization_user_updated(self):
        default_org = Organization.objects.first()
        org_user = self._create_staff_org_admin()
        self.assertNotEqual(org_user.organization_id, default_org.pk)
        self.assertEqual(ns_queryset.count(), 1)
        org_user.organization_id = default_org.pk
        org_user.full_clean()
        org_user.save()

        self.assertEqual(ns_queryset.filter(deleted=True).count(), 1)
        self.assertEqual(ns_queryset.filter(deleted=False).count(), 1)
        notification_setting = ns_queryset.first()
        self.assertEqual(notification_setting.organization.pk, default_org.pk)

    def test_organization_user_no_change_save(self):
        org_user = self._create_staff_org_admin()
        ns = ns_queryset.first()
        self.assertEqual(ns_queryset.count(), 1)
        org_user.full_clean()
        org_user.save()
        self.assertEqual(ns_queryset.count(), 1)
        update_ns = ns_queryset.first()
        self.assertEqual(ns.id, update_ns.id)

    def test_org_user_promoted_to_org_admin_with_org_change(self):
        default_org = Organization.objects.get(slug='default')
        org_user = self._create_org_user(user=self._create_operator(), is_admin=False)
        self.assertEqual(ns_queryset.count(), 0)
        org_user.is_admin = True
        org_user.organization = default_org
        org_user.full_clean()
        org_user.save()
        self.assertEqual(ns_queryset.count(), 1)
        ns = ns_queryset.first()
        self.assertEqual(ns.organization, default_org)
        self.assertEqual(ns.user, org_user.user)

    def test_deleted_notificationsetting_autocreated(self):
        org_user = self._create_staff_org_admin()
        self.assertEqual(ns_queryset.count(), 1)
        ns = ns_queryset.first()
        ns.deleted = True
        ns.full_clean()
        ns.save()

        # Emit post_save for organization user
        post_save.send(sender=OrganizationUser, instance=org_user, created=False)

        self.assertEqual(ns_queryset.count(), 1)
        ns.refresh_from_db()
        self.assertEqual(ns.deleted, False)



from django.conf import settings
from django.contrib.sites.models import Site
from django.urls import NoReverseMatch, reverse


def _get_object_link(obj, field, absolute_url=False, *args, **kwargs):
    related_obj = getattr(obj, field)
    try:
        url = reverse(
            f'admin:{related_obj._meta.app_label}_{related_obj._meta.model_name}_change',
            args=[related_obj.id],
        )
        if absolute_url:
            url = _get_absolute_url(url)
        return url
    except (NoReverseMatch, AttributeError):
        return '#'


def _get_absolute_url(path):
    site = Site.objects.get_current()
    protocol = 'http' if getattr(settings, 'DEBUG', False) else 'https'
    return f'{protocol}://{site.domain}{path}'


def normalize_unread_count(unread_count):
    if unread_count > 99:
        return '99+'
    else:
        return unread_count



from django.db import migrations, models

from openwisp_notifications.types import NOTIFICATION_CHOICES


class Migration(migrations.Migration):
    dependencies = [
        ('openwisp_notifications', '0002_default_permissions'),
    ]

    operations = [
        migrations.AddField(
            model_name='notification',
            name='type',
            field=models.CharField(
                choices=NOTIFICATION_CHOICES,
                max_length=30,
                null=True,
            ),
        ),
    ]



from django.contrib.staticfiles.testing import StaticLiveServerTestCase
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

from openwisp_notifications.signals import notify
from openwisp_notifications.swapper import load_model
from openwisp_notifications.utils import _get_object_link
from openwisp_users.tests.utils import TestOrganizationMixin
from openwisp_utils.test_selenium_mixins import SeleniumTestMixin

Notification = load_model('Notification')


class TestWidget(
    SeleniumTestMixin,
    TestOrganizationMixin,
    StaticLiveServerTestCase,
):
    serve_static = True

    def setUp(self):
        self.admin = self._create_admin(
            username=self.admin_username, password=self.admin_password
        )
        self.operator = super()._get_operator()
        self.notification_options = dict(
            sender=self.admin,
            recipient=self.admin,
            verb='Test Notification',
            email_subject='Test Email subject',
            action_object=self.operator,
            target=self.operator,
            type='default',
        )

    def _create_notification(self):
        return notify.send(**self.notification_options)

    def test_notification_relative_link(self):
        self.login()
        notification = self._create_notification().pop()[1][0]
        self.web_driver.find_element(By.ID, 'openwisp_notifications').click()
        WebDriverWait(self.web_driver, 10).until(
            EC.visibility_of_element_located((By.CLASS_NAME, 'ow-notification-elem'))
        )
        notification_elem = self.web_driver.find_element(
            By.CLASS_NAME, 'ow-notification-elem'
        )
        data_location_value = notification_elem.get_attribute('data-location')
        self.assertEqual(
            data_location_value, _get_object_link(notification, 'target', False)
        )

    def test_notification_dialog(self):
        self.login()
        self.notification_options.update(
            {'message': 'Test Message', 'description': 'Test Description'}
        )
        notification = self._create_notification().pop()[1][0]
        self.web_driver.find_element(By.ID, 'openwisp_notifications').click()
        WebDriverWait(self.web_driver, 10).until(
            EC.visibility_of_element_located((By.ID, f'ow-{notification.id}'))
        )
        self.web_driver.find_element(By.ID, f'ow-{notification.id}').click()
        WebDriverWait(self.web_driver, 10).until(
            EC.visibility_of_element_located((By.CLASS_NAME, 'ow-dialog-notification'))
        )
        dialog = self.web_driver.find_element(By.CLASS_NAME, 'ow-dialog-notification')
        self.assertEqual(
            dialog.find_element(By.CLASS_NAME, 'ow-message-title').text, 'Test Message'
        )
        self.assertEqual(
            dialog.find_element(By.CLASS_NAME, 'ow-message-description').text,
            'Test Description',
        )

    def test_notification_dialog_open_button_visibility(self):
        self.login()
        self.notification_options.pop('target')
        self.notification_options.update(
            {'message': 'Test Message', 'description': 'Test Description'}
        )
        notification = self._create_notification().pop()[1][0]
        self.web_driver.find_element(By.ID, 'openwisp_notifications').click()
        WebDriverWait(self.web_driver, 10).until(
            EC.visibility_of_element_located((By.ID, f'ow-{notification.id}'))
        )
        self.web_driver.find_element(By.ID, f'ow-{notification.id}').click()
        WebDriverWait(self.web_driver, 10).until(
            EC.visibility_of_element_located((By.CLASS_NAME, 'ow-dialog-notification'))
        )
        dialog = self.web_driver.find_element(By.CLASS_NAME, 'ow-dialog-notification')
        # This confirms the button is hidden
        dialog.find_element(By.CSS_SELECTOR, '.ow-message-target-redirect.ow-hide')



# Generated by Django 3.0.5 on 2020-04-14 18:32

import uuid

import django.db.models.deletion
import django.utils.timezone
import jsonfield.fields
import model_utils.fields
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='NotificationUser',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    'created',
                    model_utils.fields.AutoCreatedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name='created',
                    ),
                ),
                (
                    'modified',
                    model_utils.fields.AutoLastModifiedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name='modified',
                    ),
                ),
                (
                    'receive',
                    models.BooleanField(
                        default=True,
                        help_text=(
                            'note: non-superadmin users receive notifications only '
                            'for organizations of which they are member of.'
                        ),
                        verbose_name='receive notifications',
                    ),
                ),
                (
                    'email',
                    models.BooleanField(
                        default=True,
                        help_text=(
                            'note: non-superadmin users receive notifications only '
                            'for organizations of which they are member of.'
                        ),
                        verbose_name='email notifications',
                    ),
                ),
                (
                    'user',
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                'verbose_name': 'user notification settings',
                'verbose_name_plural': 'user notification settings',
            },
        ),
        migrations.CreateModel(
            name='Notification',
            fields=[
                (
                    'level',
                    models.CharField(
                        choices=[
                            ('success', 'success'),
                            ('info', 'info'),
                            ('warning', 'warning'),
                            ('error', 'error'),
                        ],
                        default='info',
                        max_length=20,
                        verbose_name='level',
                    ),
                ),
                (
                    'unread',
                    models.BooleanField(
                        db_index=True, default=True, verbose_name='unread'
                    ),
                ),
                (
                    'actor_object_id',
                    models.CharField(max_length=255, verbose_name='actor object id'),
                ),
                ('verb', models.CharField(max_length=255, verbose_name='verb')),
                (
                    'description',
                    models.TextField(blank=True, null=True, verbose_name='description'),
                ),
                (
                    'target_object_id',
                    models.CharField(
                        blank=True,
                        max_length=255,
                        null=True,
                        verbose_name='target object id',
                    ),
                ),
                (
                    'action_object_object_id',
                    models.CharField(
                        blank=True,
                        max_length=255,
                        null=True,
                        verbose_name='action object object id',
                    ),
                ),
                (
                    'timestamp',
                    models.DateTimeField(
                        db_index=True,
                        default=django.utils.timezone.now,
                        verbose_name='timestamp',
                    ),
                ),
                (
                    'public',
                    models.BooleanField(
                        db_index=True, default=True, verbose_name='public'
                    ),
                ),
                (
                    'deleted',
                    models.BooleanField(
                        db_index=True, default=False, verbose_name='deleted'
                    ),
                ),
                (
                    'emailed',
                    models.BooleanField(
                        db_index=True, default=False, verbose_name='emailed'
                    ),
                ),
                (
                    'data',
                    jsonfield.fields.JSONField(
                        blank=True, null=True, verbose_name='data'
                    ),
                ),
                (
                    'id',
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    'action_object_content_type',
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='notify_action_object',
                        to='contenttypes.ContentType',
                        verbose_name='action object content type',
                    ),
                ),
                (
                    'actor_content_type',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='notify_actor',
                        to='contenttypes.ContentType',
                        verbose_name='actor content type',
                    ),
                ),
                (
                    'recipient',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='notifications',
                        to=settings.AUTH_USER_MODEL,
                        verbose_name='recipient',
                    ),
                ),
                (
                    'target_content_type',
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='notify_target',
                        to='contenttypes.ContentType',
                        verbose_name='target content type',
                    ),
                ),
            ],
            options={
                'ordering': ('-timestamp',),
                'abstract': False,
                'swappable': 'OPENWISP_NOTIFICATIONS_NOTIFICATION_MODEL',
                'index_together': {('recipient', 'unread')},
                'verbose_name': 'Notification',
                'verbose_name_plural': 'Notifications',
            },
        ),
    ]



# Generated by Django 3.1.13 on 2021-09-17 11:09

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('openwisp_notifications', '0006_objectnotification'),
    ]

    operations = [
        migrations.AddField(
            model_name='notificationsetting',
            name='deleted',
            field=models.BooleanField(
                blank=True, default=False, null=True, verbose_name='Delete'
            ),
        ),
    ]



import sys
import uuid
from datetime import timedelta
from unittest.mock import patch

import pytest
from channels.db import database_sync_to_async
from channels.testing import WebsocketCommunicator
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.core.cache import cache
from django.utils.module_loading import import_string
from django.utils.timezone import now
from freezegun import freeze_time

from openwisp_notifications.api.serializers import NotificationListSerializer
from openwisp_notifications.signals import notify
from openwisp_notifications.swapper import load_model
from openwisp_notifications.tests.test_helpers import TEST_DATETIME
from openwisp_notifications.websockets.consumers import NotificationConsumer

User = get_user_model()
Notification = load_model('Notification')
IgnoreObjectNotification = load_model('IgnoreObjectNotification')


@database_sync_to_async
def create_notification(admin_user):
    n = notify.send(sender=admin_user, type='default').pop()
    return n[1].pop()


@database_sync_to_async
def bulk_create_notification(admin_user, count=10):
    notifications = []
    for _ in range(count):
        notifications.append(
            Notification(
                recipient=admin_user,
                actor=admin_user,
                type='default',
            )
        )
    Notification.objects.bulk_create(notifications)


@database_sync_to_async
def notification_operation(notification, **kwargs):
    if kwargs.get('mark_read', False):
        notification.mark_as_read()
    if kwargs.get('delete', False):
        notification.delete()
    if kwargs.get('refresh', False):
        notification.refresh_from_db()
    return Notification.objects.first()


@database_sync_to_async
def create_object_notification(admin_user):
    obj = IgnoreObjectNotification.objects.create(
        object_id=admin_user.pk,
        object_content_type_id=ContentType.objects.get_for_model(
            admin_user._meta.model
        ).pk,
        user_id=admin_user.pk,
        valid_till=TEST_DATETIME,
    )
    return obj


@pytest.mark.asyncio
@pytest.mark.django_db(transaction=True)
class TestNotificationSockets:
    application = import_string(getattr(settings, 'ASGI_APPLICATION'))

    async def _get_communicator(self, admin_client):
        session_id = admin_client.cookies['sessionid'].value
        communicator = WebsocketCommunicator(
            self.application,
            path='ws/notification/',
            headers=[
                (
                    b'cookie',
                    f'sessionid={session_id}'.encode('ascii'),
                )
            ],
        )
        connected, subprotocol = await communicator.connect()
        assert connected is True
        return communicator

    async def test_new_notification_created(self, admin_user, admin_client):
        communicator = await self._get_communicator(admin_client)
        n = await create_notification(admin_user)
        response = await communicator.receive_json_from()
        expected_response = {
            'type': 'notification',
            'notification_count': 1,
            'reload_widget': True,
            'notification': NotificationListSerializer(n).data,
        }
        assert response == expected_response
        await communicator.disconnect()

    async def test_read_notification(self, admin_user, admin_client):
        n = await create_notification(admin_user)
        communicator = await self._get_communicator(admin_client)
        await notification_operation(n, mark_read=True)
        response = await communicator.receive_json_from(timeout=2)
        expected_response = {
            'type': 'notification',
            'notification_count': 0,
            'reload_widget': False,
            'notification': None,
        }
        assert response == expected_response
        await communicator.disconnect()

    async def test_delete_notification(self, admin_user, admin_client):
        n = await create_notification(admin_user)
        communicator = await self._get_communicator(admin_client)
        await notification_operation(n, delete=True)
        response = await communicator.receive_json_from()
        expected_response = {
            'type': 'notification',
            'notification_count': 0,
            'reload_widget': True,
            'notification': None,
        }
        assert response == expected_response
        await communicator.disconnect()

    async def test_unauthenticated_user(self, client):
        client.cookies['sessionid'] = 'random'
        with pytest.raises(AssertionError):
            await self._get_communicator(client)

    async def test_receive_notification_data(self, admin_user, admin_client):
        communicator = await self._get_communicator(admin_client)
        n = await create_notification(admin_user)
        # Assert message for new notification created
        res = await communicator.receive_json_from()
        assert res['notification_count'] == 1
        await communicator.send_json_to(
            {'type': 'notification', 'notification_id': str(n.id)}
        )
        # Assert message for notification mark read
        res = await communicator.receive_json_from()
        assert res['notification_count'] == 0
        assert await communicator.receive_nothing() is True
        await communicator.disconnect()

    async def test_receive_with_improper_data(self, admin_user, admin_client):
        communicator = await self._get_communicator(admin_client)
        # Check for JSONDecodeError
        await communicator.send_to('Not JSON')
        assert await communicator.receive_nothing() is True

        await communicator.send_json_to(
            {'type': 'notification', 'notification': 'random'}
        )
        assert await communicator.receive_nothing() is True

        await communicator.send_json_to(
            {'type': 'notification', 'notification_id': str(uuid.uuid4())}
        )
        assert await communicator.receive_nothing() is True
        await communicator.disconnect()

    async def test_retreive_object_notification_valid_till(
        self, admin_user, admin_client
    ):
        communicator = await self._get_communicator(admin_client)
        # Test for non-existing object
        payload = {
            'type': 'object_notification',
            'app_label': admin_user._meta.app_label,
            'model_name': admin_user._meta.model_name,
            'object_id': str(admin_user.pk),
        }
        await communicator.send_json_to(payload)
        assert await communicator.receive_nothing() is True

        # Test for an existing object
        with patch('openwisp_notifications.tasks.delete_ignore_object_notification'):
            await create_object_notification(admin_user)
            await communicator.send_json_to(payload)
            response = await communicator.receive_json_from()
            assert response['type'] == 'object_notification'
            assert response['valid_till'].split('T')[0] in TEST_DATETIME.isoformat()
        await communicator.disconnect()

    async def test_short_term_notification_storm_prevention(
        self, admin_user, admin_client
    ):
        communicator = await self._get_communicator(admin_client)
        for _ in range(6):
            await create_notification(admin_user)
            response = await communicator.receive_json_from()
            assert response['notification'] is not None
            assert response['reload_widget'] is True
        # After notification storms prevention starts
        for _ in range(4):
            await create_notification(admin_user)
            response = await communicator.receive_json_from()
            assert response['notification'] is None
            assert response['reload_widget'] is False

        assert await communicator.receive_nothing() is True
        await communicator.disconnect()

    @patch.object(NotificationConsumer, '_backoff_increment', 60)
    @patch.object(NotificationConsumer, '_max_allowed_backoff', 5)
    async def test_long_term_notification_storm_prevention(
        self, admin_user, admin_client
    ):
        if sys.version_info[:2] == (3, 7):
            NotificationConsumer._backoff_increment = 60
            NotificationConsumer._max_allowed_backoff = 5

        datetime_now = now()
        freezer = freeze_time(datetime_now).start()
        await bulk_create_notification(admin_user, count=30)

        # Establish connection
        communicator = await self._get_communicator(admin_client)

        # Long term notification storm prevention takes over
        freezer.move_to(datetime_now + timedelta(seconds=30))
        # Set notification storm prevention in action.
        # _initial_backoff is mocked since the time is frozen in this test
        # which was leading to initialization case getting skipped.
        with patch.object(NotificationConsumer, '_initial_backoff', 0):
            if sys.version_info[:2] == (3, 7):
                NotificationConsumer._initial_backoff = 0
            await create_notification(admin_user)
            response = await communicator.receive_json_from()
            if sys.version_info[:2] == (3, 7):
                NotificationConsumer._initial_backoff = 1

        # These notifications will increment the backoff time beyond allowed limit.
        for _ in range(2):
            await create_notification(admin_user)
            response = await communicator.receive_json_from()
            assert response['notification'] is None
            assert response['reload_widget'] is False

        # Reload widget when maximum backoff limit is reached
        freezer.move_to(datetime_now + timedelta(seconds=40))
        await create_notification(admin_user)
        response = await communicator.receive_json_from()
        assert response['notification'] is None
        assert response['reload_widget'] is True

        # user_in_notification_storm cache expires and
        # notification storm ends.
        cache.delete(f'ow-noti-storm-{admin_user.pk}')
        freezer.move_to(datetime_now + timedelta(seconds=200))
        await create_notification(admin_user)
        response = await communicator.receive_json_from()
        assert response['notification'] is not None
        assert response['reload_widget'] is True

        await communicator.disconnect()

        if sys.version_info[:1] == (3, 7):
            NotificationConsumer._backoff_increment = 1
            NotificationConsumer._max_allowed_backoff = 15



# -*- coding: utf-8 -*-
from django.core.cache import cache
from django.template import Library
from django.utils.html import format_html

from openwisp_notifications.swapper import load_model
from openwisp_notifications.utils import normalize_unread_count

Notification = load_model('Notification')

register = Library()


def _get_user_unread_count(user):
    return user.notifications.unread().count()


def get_notifications_count(context):
    user_pk = context['user'].is_authenticated and context['user'].pk
    cache_key = Notification.count_cache_key(user_pk)
    count = cache.get(cache_key)
    if count is None:
        count = _get_user_unread_count(context['user'])
        count = normalize_unread_count(count)
        cache.set(cache_key, count)
    return count


def unread_notifications(context):
    count = get_notifications_count(context)
    output = ''
    if count:
        output = '<span id="ow-notification-count">{0}</span>'
        output = format_html(output.format(count))
    return output


register.simple_tag(takes_context=True)(unread_notifications)



from django.core.management.base import BaseCommand

from openwisp_notifications.handlers import (
    notification_type_registered_unregistered_handler,
)
from openwisp_notifications.signals import notify
from openwisp_notifications.swapper import swapper_load_model

Organization = swapper_load_model('openwisp_users', 'Organization')


class BaseCreateNotificationCommand(BaseCommand):
    def handle(self, *args, **kwargs):
        default_org = Organization.objects.first()
        default_org.organization_id = default_org.id
        notify.send(sender=default_org, type='default', target=default_org)


class BasePopulateNotificationPreferencesCommand(BaseCommand):
    def handle(self, *args, **kwargs):
        notification_type_registered_unregistered_handler(self)



# Generated by Django 3.0.8 on 2020-07-23 14:20

import uuid

import django.db.models.deletion
import swapper
from django.conf import settings
from django.contrib.auth.management import create_permissions
from django.db import migrations, models

from openwisp_notifications.migrations import get_swapped_model
from openwisp_notifications.types import NOTIFICATION_CHOICES


def create_notification_setting_groups_permissions(apps, schema_editor):
    # Populate permissions
    app_config = apps.get_app_config('openwisp_notifications')
    app_config.models_module = True
    create_permissions(app_config, apps=apps, verbosity=0)
    app_config.models_module = None

    Group = get_swapped_model(apps, 'openwisp_users', 'Group')
    Permission = apps.get_model('auth', 'Permission')

    operator = Group.objects.filter(name='Operator')
    if operator.count() == 0:
        operator = Group.objects.create(name='Operator')
    else:
        operator = operator.first()

    admin = Group.objects.filter(name='Administrator')
    if admin.count() == 0:
        admin = Group.objects.create(name='Administrator')
    else:
        admin = admin.first()

    permissions = [
        Permission.objects.get(
            content_type__app_label='openwisp_notifications',
            content_type__model='notificationsetting',
            codename='change_notificationsetting',
        ).pk,
    ]

    permissions += operator.permissions.all()
    operator.permissions.set(permissions)

    permissions += admin.permissions.all()
    admin.permissions.set(permissions)


def reverse_notification_setting_groups_permissions(apps, schema_editor):
    Group = get_swapped_model(apps, 'openwisp_users', 'Group')
    operator = Group.objects.filter(name='Operator').first()
    administrator = Group.objects.filter(name='Administrator').first()

    if operator is not None:
        operator.permissions.filter(
            content_type__app_label='openwisp_notifications',
            content_type__model='notificationsetting',
        ).delete()
    if administrator is not None:
        administrator.permissions.filter(
            content_type__app_label='openwisp_notifications',
            content_type__model='notificationsetting',
        ).delete()


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        swapper.dependency('openwisp_users', 'Organization'),
        ('openwisp_notifications', '0003_notification_notification_type'),
    ]

    operations = [
        migrations.CreateModel(
            name='NotificationSetting',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    'type',
                    models.CharField(
                        choices=NOTIFICATION_CHOICES,
                        max_length=30,
                        null=True,
                        verbose_name='Notification Type',
                    ),
                ),
                (
                    'web',
                    models.BooleanField(
                        null=True,
                        blank=True,
                        help_text=(
                            'Note: Non-superadmin users receive notifications only '
                            'for organizations of which they are member of.'
                        ),
                        verbose_name='web notifications',
                    ),
                ),
                (
                    'email',
                    models.BooleanField(
                        null=True,
                        blank=True,
                        help_text=(
                            'Note: Non-superadmin users receive notifications only '
                            'for organizations of which they are member of.'
                        ),
                        verbose_name='email notifications',
                    ),
                ),
                (
                    'organization',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=swapper.get_model_name('openwisp_users', 'Organization'),
                    ),
                ),
                (
                    'user',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                'verbose_name': 'user notification settings',
                'verbose_name_plural': 'user notification settings',
                'ordering': ['organization', 'type'],
                'abstract': False,
                'swappable': 'OPENWISP_NOTIFICATIONS_NOTIFICATIONSETTING_MODEL',
            },
        ),
        migrations.AddConstraint(
            model_name='notificationsetting',
            constraint=models.UniqueConstraint(
                fields=('organization', 'type', 'user'),
                name='unique_notification_setting',
            ),
        ),
        migrations.AddIndex(
            model_name='notificationsetting',
            index=models.Index(
                fields=['type', 'organization'], name='openwisp_no_type_5a6a77_idx'
            ),
        ),
        migrations.RunPython(
            create_notification_setting_groups_permissions,
            reverse_code=reverse_notification_setting_groups_permissions,
        ),
    ]



import json

from asgiref.sync import async_to_sync
from channels.generic.websocket import WebsocketConsumer
from django.contrib.contenttypes.models import ContentType
from django.utils.timezone import now, timedelta

from openwisp_notifications.api.serializers import IgnoreObjectNotificationSerializer
from openwisp_notifications.swapper import load_model

from .. import settings as app_settings

Notification = load_model('Notification')
IgnoreObjectNotification = load_model('IgnoreObjectNotification')


class NotificationConsumer(WebsocketConsumer):
    _initial_backoff = app_settings.NOTIFICATION_STORM_PREVENTION['initial_backoff']
    _backoff_increment = app_settings.NOTIFICATION_STORM_PREVENTION['backoff_increment']
    _max_allowed_backoff = app_settings.NOTIFICATION_STORM_PREVENTION[
        'max_allowed_backoff'
    ]

    def _is_user_authenticated(self):
        try:
            assert self.scope['user'].is_authenticated is True
        except (KeyError, AssertionError):
            self.close()
            return False
        else:
            return True

    def connect(self):
        if self._is_user_authenticated():
            async_to_sync(self.channel_layer.group_add)(
                'ow-notification-{0}'.format(self.scope['user'].pk), self.channel_name
            )
            self.accept()
            self.scope['last_update_datetime'] = now()
            self.scope['backoff'] = self._initial_backoff

    def disconnect(self, close_code):
        async_to_sync(self.channel_layer.group_discard)(
            'ow-notification-{0}'.format(self.scope['user'].pk), self.channel_name
        )

    def process_event_for_notification_storm(self, event):
        if not event['in_notification_storm']:
            return event
        # Removing notification is required to prevent frontend
        # from showing toasts.
        event['notification'] = None
        datetime_now = now()
        # If delay exceeds max_allowed_backoff, reset and send
        # update. This is required to trigger reloading of
        # notification widget.
        if (
            self.scope['last_update_datetime'] > datetime_now
            and (self.scope['last_update_datetime'] - datetime_now).seconds
            >= self._max_allowed_backoff
        ):
            self.scope['last_update_datetime'] = datetime_now
            self.scope['backoff'] = self._initial_backoff
            event['notification'] = None
        elif self.scope['last_update_datetime'] > datetime_now - timedelta(
            seconds=self._initial_backoff
        ):
            self.scope['last_update_datetime'] = datetime_now + timedelta(
                seconds=self.scope['backoff']
            )
            self.scope['backoff'] = self.scope['backoff'] + self._backoff_increment
            event['reload_widget'] = False
        else:
            self.scope['last_update_datetime'] = datetime_now
            self.scope['backoff'] = self._initial_backoff
        return event

    def send_updates(self, event):
        event = self.process_event_for_notification_storm(event)
        self.send(
            json.dumps(
                {
                    'type': 'notification',
                    'notification_count': event['notification_count'],
                    'reload_widget': event['reload_widget'],
                    'notification': event['notification'],
                }
            )
        )

    def receive(self, text_data):
        if self._is_user_authenticated():
            try:
                json_data = json.loads(text_data)
            except json.JSONDecodeError:
                return

            try:
                if json_data['type'] == 'notification':
                    self._notification_handler(
                        notification_id=json_data['notification_id']
                    )
                elif json_data['type'] == 'object_notification':
                    self._object_notification_handler(
                        object_id=json_data['object_id'],
                        app_label=json_data['app_label'],
                        model_name=json_data['model_name'],
                    )
            except KeyError:
                return

    def _notification_handler(self, notification_id):
        try:
            notification = Notification.objects.get(
                recipient=self.scope['user'], id=notification_id
            )
            notification.mark_as_read()
        except Notification.DoesNotExist:
            return

    def _object_notification_handler(self, object_id, app_label, model_name):
        try:
            object_notification = IgnoreObjectNotification.objects.get(
                user=self.scope['user'],
                object_id=object_id,
                object_content_type_id=ContentType.objects.get_by_natural_key(
                    app_label=app_label,
                    model=model_name,
                ).pk,
            )
            serialized_data = IgnoreObjectNotificationSerializer(object_notification)
            self.send(
                json.dumps(
                    {
                        'type': 'object_notification',
                        'valid_till': serialized_data.data['valid_till'],
                    }
                )
            )
        except IgnoreObjectNotification.DoesNotExist:
            return



from django.contrib.contenttypes.models import ContentType
from django.http import Http404, HttpResponseRedirect
from django.urls import reverse
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import serializers, status
from rest_framework.authentication import SessionAuthentication
from rest_framework.generics import (
    GenericAPIView,
    RetrieveDestroyAPIView,
    RetrieveUpdateAPIView,
    get_object_or_404,
)
from rest_framework.mixins import CreateModelMixin, ListModelMixin, UpdateModelMixin
from rest_framework.pagination import PageNumberPagination
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response

from openwisp_notifications.api.serializers import (
    IgnoreObjectNotificationSerializer,
    NotificationListSerializer,
    NotificationSerializer,
    NotificationSettingSerializer,
)
from openwisp_notifications.swapper import load_model
from openwisp_users.api.authentication import BearerAuthentication

from .filters import NotificationSettingFilter

UNAUTHORIZED_STATUS_CODES = (
    status.HTTP_401_UNAUTHORIZED,
    status.HTTP_403_FORBIDDEN,
)

Notification = load_model('Notification')
NotificationSetting = load_model('NotificationSetting')
IgnoreObjectNotification = load_model('IgnoreObjectNotification')


class NotificationPaginator(PageNumberPagination):
    page_size = 20
    page_size_query_param = 'page_size'
    max_page_size = 100


class BaseNotificationView(GenericAPIView):
    model = Notification
    authentication_classes = [BearerAuthentication, SessionAuthentication]
    permission_classes = [IsAuthenticated]
    queryset = Notification.objects.all()
    # NOTE: Required for Browsable API view. Don't remove.
    serializer_class = serializers.Serializer

    def get_queryset(self):
        if getattr(self, 'swagger_fake_view', False):
            return self.queryset.none()  # pragma: no cover
        return self.queryset.filter(recipient=self.request.user)


class NotificationListView(BaseNotificationView, ListModelMixin):
    serializer_class = NotificationListSerializer
    pagination_class = NotificationPaginator
    filter_backends = [DjangoFilterBackend]
    filterset_fields = ['unread']

    def get(self, request, *args, **kwargs):
        return self.list(request, *args, **kwargs)


class NotificationDetailView(BaseNotificationView, RetrieveDestroyAPIView):
    serializer_class = NotificationSerializer
    lookup_field = 'pk'

    def patch(self, request, *args, **kwargs):
        return self._mark_notification_read()

    def _mark_notification_read(self):
        notification = self.get_object()
        notification.mark_as_read()
        return Response(
            status=status.HTTP_200_OK,
        )


class NotificationReadRedirect(BaseNotificationView):
    lookup_field = 'pk'

    def get(self, request, *args, **kwargs):
        notification = self.get_object()
        notification.mark_as_read()
        return HttpResponseRedirect(notification.target_url)

    def handle_exception(self, exc):
        response = super().handle_exception(exc)
        if response.status_code not in UNAUTHORIZED_STATUS_CODES:
            return response

        redirect_url = '{admin_login}?next={path}'.format(
            admin_login=reverse('admin:login'), path=self.request.path
        )
        return HttpResponseRedirect(redirect_url)


class NotificationReadAllView(BaseNotificationView):
    def post(self, request, *args, **kwargs):
        queryset = self.get_queryset()
        queryset.filter(unread=True).update(unread=False)
        # update() does not create post_save signal, therefore
        # manual invalidation of cache is required
        Notification.invalidate_unread_cache(request.user)
        return Response(status=status.HTTP_200_OK)


class BaseNotificationSettingView(GenericAPIView):
    model = NotificationSetting
    serializer_class = NotificationSettingSerializer
    authentication_classes = [BearerAuthentication, SessionAuthentication]
    permission_classes = [IsAuthenticated]

    def get_queryset(self):
        if getattr(self, 'swagger_fake_view', False):
            return NotificationSetting.objects.none()  # pragma: no cover
        return NotificationSetting.objects.filter(user=self.request.user)


class NotificationSettingListView(BaseNotificationSettingView, ListModelMixin):
    pagination_class = NotificationPaginator
    filter_backends = [DjangoFilterBackend]
    filterset_class = NotificationSettingFilter

    def get(self, request, *args, **kwargs):
        return self.list(request, *args, **kwargs)


class NotificationSettingView(BaseNotificationSettingView, RetrieveUpdateAPIView):
    lookup_field = 'pk'


class BaseIgnoreObjectNotificationView(GenericAPIView):
    model = IgnoreObjectNotification
    serializer_class = IgnoreObjectNotificationSerializer
    authentication_classes = [BearerAuthentication, SessionAuthentication]
    permission_classes = [IsAuthenticated]

    def get_queryset(self):
        return IgnoreObjectNotification.objects.filter(user=self.request.user)


class IgnoreObjectNotificationListView(
    BaseIgnoreObjectNotificationView, ListModelMixin
):
    pagination_class = NotificationPaginator

    def get(self, request, *args, **kwargs):
        return self.list(request, *args, **kwargs)


class IgnoreObjectNotificationView(
    BaseIgnoreObjectNotificationView,
    RetrieveDestroyAPIView,
    CreateModelMixin,
    UpdateModelMixin,
):
    def initial(self, request, *args, **kwargs):
        super().initial(request, *args, *kwargs)
        self.kwargs['obj_content_type_id'] = self.obj_content_type_id

    def put(self, request, *args, **kwargs):
        return self.create_or_update(request, *args, **kwargs)

    @property
    def obj_content_type_id(self):
        return ContentType.objects.get_by_natural_key(
            self.kwargs['app_label'], self.kwargs['model_name']
        ).pk

    def get_object(self):
        queryset = self.filter_queryset(self.get_queryset())
        filter_kwargs = {
            'object_content_type_id': self.kwargs['obj_content_type_id'],
            'object_id': self.kwargs['object_id'],
        }
        obj = get_object_or_404(queryset, **filter_kwargs)
        # May raise a permission denied
        self.check_object_permissions(self.request, obj)
        return obj

    def create_or_update(self, request, *args, **kwargs):
        try:
            return super().update(request, *args, **kwargs)
        except Http404:
            return super().create(request, *args, **kwargs)

    def perform_create(self, serializer):
        serializer.save(
            user=self.request.user,
            object_content_type_id=self.kwargs['obj_content_type_id'],
            object_id=self.kwargs['object_id'],
        )


notifications_list = NotificationListView.as_view()
notification_detail = NotificationDetailView.as_view()
notifications_read_all = NotificationReadAllView.as_view()
notification_read_redirect = NotificationReadRedirect.as_view()
notification_setting_list = NotificationSettingListView.as_view()
notification_setting = NotificationSettingView.as_view()
ignore_object_notification_list = IgnoreObjectNotificationListView.as_view()
ignore_object_notification = IgnoreObjectNotificationView.as_view()



from datetime import timedelta
from unittest.mock import patch

from allauth.account.models import EmailAddress
from celery.exceptions import OperationalError
from django.apps.registry import apps
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.core import mail
from django.core.cache import cache
from django.core.exceptions import ImproperlyConfigured
from django.db.models.signals import post_migrate, post_save
from django.template import TemplateDoesNotExist
from django.test import TransactionTestCase
from django.urls import reverse
from django.utils import timezone
from django.utils.timesince import timesince

from openwisp_notifications import settings as app_settings
from openwisp_notifications import tasks
from openwisp_notifications.exceptions import NotificationRenderException
from openwisp_notifications.handlers import (
    notify_handler,
    register_notification_cache_update,
)
from openwisp_notifications.signals import notify
from openwisp_notifications.swapper import load_model, swapper_load_model
from openwisp_notifications.tests.test_helpers import (
    mock_notification_types,
    register_notification_type,
    unregister_notification_type,
)
from openwisp_notifications.types import (
    _unregister_notification_choice,
    get_notification_configuration,
)
from openwisp_notifications.utils import _get_absolute_url
from openwisp_users.tests.utils import TestOrganizationMixin
from openwisp_utils.tests import capture_any_output

User = get_user_model()

Notification = load_model('Notification')
NotificationSetting = load_model('NotificationSetting')
NotificationAppConfig = apps.get_app_config(Notification._meta.app_label)


OrganizationUser = swapper_load_model('openwisp_users', 'OrganizationUser')
Group = swapper_load_model('openwisp_users', 'Group')

start_time = timezone.now()
ten_minutes_ago = start_time - timedelta(minutes=10)
notification_queryset = Notification.objects.order_by('-timestamp')


class TestNotifications(TestOrganizationMixin, TransactionTestCase):
    app_label = 'openwisp_notifications'

    def setUp(self):
        self.admin = self._create_admin()
        self.notification_options = dict(
            sender=self.admin,
            description='Test Notification',
            verb='Test Notification',
            email_subject='Test Email subject',
            url='https://localhost:8000/admin',
        )

    def _create_notification(self):
        return notify.send(**self.notification_options)

    def test_create_notification(self):
        operator = super()._create_operator()
        data = dict(
            email_subject='Test Email subject', url='https://localhost:8000/admin'
        )
        n = Notification.objects.create(
            actor=self.admin,
            recipient=self.admin,
            description='Test Notification Description',
            verb='Test Notification',
            action_object=operator,
            target=operator,
            data=data,
        )
        self.assertEqual(str(n), timesince(n.timestamp, timezone.now()))
        self.assertEqual(n.actor_object_id, self.admin.id)
        self.assertEqual(
            n.actor_content_type, ContentType.objects.get_for_model(self.admin)
        )
        self.assertEqual(n.action_object_object_id, operator.id)
        self.assertEqual(
            n.action_object_content_type, ContentType.objects.get_for_model(operator)
        )
        self.assertEqual(n.target_object_id, operator.id)
        self.assertEqual(
            n.target_content_type, ContentType.objects.get_for_model(operator)
        )
        self.assertEqual(n.verb, 'Test Notification')
        self.assertEqual(n.message, 'Test Notification Description')
        self.assertEqual(n.recipient, self.admin)

    def test_create_with_extra_data(self):
        register_notification_type(
            'error_type',
            {
                'verbose_name': 'Error',
                'level': 'error',
                'verb': 'error',
                'message': 'Error: {error}',
                'email_subject': 'Error subject: {error}',
            },
        )
        error = '500 Internal Server Error'
        notify.send(
            type='error_type',
            url='https://localhost:8000/admin',
            recipient=self.admin,
            sender=self.admin,
            error=error,
        )
        self.assertEqual(notification_queryset.count(), 1)
        n = notification_queryset.first()
        self.assertIn(f'Error: {error}', n.message)
        self.assertEqual(n.email_subject, f'Error subject: {error}')

    def test_superuser_notifications_disabled(self):
        target_obj = self._get_org_user()
        self.notification_options.update({'type': 'default', 'target': target_obj})
        notification_preference = NotificationSetting.objects.get(
            user_id=self.admin.pk,
            organization_id=target_obj.organization.pk,
            type='default',
        )
        self.assertEqual(notification_preference.email, None)
        notification_preference.web = False
        notification_preference.save()
        notification_preference.refresh_from_db()
        self.assertEqual(notification_preference.email, False)
        self._create_notification()
        self.assertEqual(notification_queryset.count(), 0)

    def test_email_sent(self):
        self._create_notification()
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].to, [self.admin.email])
        n = notification_queryset.first()
        self.assertEqual(mail.outbox[0].subject, n.data.get('email_subject'))
        self.assertIn(n.message, mail.outbox[0].body)
        self.assertIn(n.data.get('url'), mail.outbox[0].body)
        self.assertIn('https://', n.data.get('url'))

    def test_email_disabled(self):
        self.notification_options.update(
            {'type': 'default', 'target': self._get_org_user()}
        )
        NotificationSetting.objects.filter(
            user_id=self.admin.pk, type='default'
        ).update(email=False)
        self._create_notification()
        self.assertEqual(Notification.objects.count(), 1)
        self.assertEqual(len(mail.outbox), 0)

    def test_notification_preference_flagged_deleted(self):
        self.notification_options.update(
            {'type': 'default', 'target': self._get_org_user()}
        )
        NotificationSetting.objects.filter(
            user_id=self.admin.pk, type='default'
        ).update(deleted=True)
        self._create_notification()
        self.assertEqual(Notification.objects.count(), 0)

    def test_notification_preference_deleted_from_db(self):
        self.notification_options.update(
            {'type': 'default', 'target': self._get_org_user()}
        )
        NotificationSetting.objects.filter(
            user_id=self.admin.pk, type='default'
        ).delete()
        self._create_notification()
        self.assertEqual(Notification.objects.count(), 0)

    def test_email_not_present(self):
        self.admin.email = ''
        self.admin.save()
        self._create_notification()
        self.assertEqual(Notification.objects.count(), 1)
        self.assertEqual(len(mail.outbox), 0)

    def test_group_recipient(self):
        org = self._get_org()
        operator = self._get_operator()
        user = self._create_user(
            username='user', email='user@user.com', first_name='User', last_name='user'
        )
        op_group = Group.objects.create(name='Operator')
        op_group.user_set.add(operator)
        op_group.user_set.add(user)
        op_group.save()
        self.notification_options.update({'recipient': op_group, 'type': 'default'})
        recipients = (operator, user)

        # Test for group with no target object
        n = self._create_notification().pop()
        if n[0] is notify_handler:
            notifications = n[1]
            self.assertEqual(len(notifications), 2)
            for notification, recipient in zip(notifications, recipients):
                self.assertEqual(notification.recipient, recipient)
        else:
            self.fail()

        # Test for group with target object of another organization
        org = self._get_org()
        target = self._create_user(
            username='target',
            email='target@target.com',
            first_name='Target',
            last_name='user',
        )
        self._create_org_user(user=target, organization=org)
        target.organization_id = org.id
        self.notification_options.update({'target': target})
        self._create_notification()
        # No new notification should be created
        self.assertEqual(notification_queryset.count(), 2)

        # Test for group with target object of same organization
        # Adding operator to organization of target object
        self._create_org_user(user=operator, organization=org, is_admin=True)
        self._create_notification()
        self.assertEqual(notification_queryset.count(), 3)
        n = notification_queryset.first()
        self.assertEqual(n.recipient, operator)

    def test_queryset_recipient(self):
        super()._create_operator()
        users = User.objects.all()
        self.notification_options.update({'recipient': users})
        n = self._create_notification().pop()
        if n[0] is notify_handler:
            notifications = n[1]
            for notification, user in zip(notifications, users):
                self.assertEqual(notification.recipient, user)
        else:
            self.fail()

    def test_description_in_email_subject(self):
        self.notification_options.pop('email_subject')
        self._create_notification()
        n = notification_queryset.first()
        self.assertEqual(mail.outbox[0].subject, n.message[0:24])

    def test_handler_optional_tag(self):
        operator = self._create_operator()
        self.notification_options.update({'action_object': operator})
        self._create_notification()
        n = notification_queryset.first()
        self.assertEqual(
            n.action_object_content_type, ContentType.objects.get_for_model(operator)
        )
        self.assertEqual(n.action_object_object_id, str(operator.id))

    def test_organization_recipient(self):
        self.notification_options.update({'type': 'default'})
        testorg = self._create_org()
        operator = self._create_operator()
        user = self._create_user(is_staff=False)
        OrganizationUser.objects.create(user=user, organization=testorg)
        OrganizationUser.objects.create(user=operator, organization=testorg)
        recipients = (self.admin, operator)
        operator.organization_id = testorg.id
        self.notification_options.update({'target': operator})
        n = self._create_notification().pop()
        if n[0] is notify_handler:
            notifications = n[1]
            for notification, recipient in zip(notifications, recipients):
                self.assertEqual(notification.recipient, recipient)
        else:
            self.fail()

    def test_no_organization(self):
        # Tests no target object is present
        self.notification_options.update({'type': 'default'})
        self._create_org_user()
        user = self._create_user(
            username='user',
            email='user@user.com',
            first_name='User',
            last_name='user',
            is_staff=False,
        )
        self._create_notification()
        # Only superadmin should receive notification
        self.assertEqual(notification_queryset.count(), 1)
        n = notification_queryset.first()
        self.assertEqual(n.actor, self.admin)
        self.assertEqual(n.recipient, self.admin)

        # Tests no user from organization of target object
        org = self._create_org(name='test_org')
        OrganizationUser.objects.create(user=user, organization=org)
        self.notification_options.update({'target': user})
        self._create_notification()
        self.assertEqual(notification_queryset.count(), 2)
        # Only superadmin should receive notification
        n = notification_queryset.first()
        self.assertEqual(n.actor, self.admin)
        self.assertEqual(n.recipient, self.admin)
        self.assertEqual(n.target, user)

    def test_default_notification_type(self):
        self.notification_options.pop('verb')
        self.notification_options.pop('url')
        self.notification_options.update(
            {'type': 'default', 'target': self._get_org_user()}
        )
        self._create_notification()
        n = notification_queryset.first()
        self.assertEqual(n.level, 'info')
        self.assertEqual(n.verb, 'default verb')
        self.assertIn(
            'Default notification with default verb and level info by', n.message
        )
        self.assertEqual(n.email_subject, '[example.com] Default Notification Subject')
        email = mail.outbox.pop()
        html_email = email.alternatives[0][0]
        self.assertEqual(
            email.body,
            (
                'Default notification with default verb and'
                ' level info by Tester Tester (test org)\n\n'
                f'For more information see {n.redirect_view_url}.'
            ),
        )
        self.assertIn(
            (
                '<div class="msg"><p>Default notification with'
                ' default verb and level info by'
                f' <a href="{n.redirect_view_url}">'
                'Tester Tester (test org)</a></p></div>'
            ),
            html_email,
        )

    def test_generic_notification_type(self):
        self.notification_options.pop('verb')
        self.notification_options.update(
            {
                'message': '[{notification.actor}]({notification.actor_link})',
                'type': 'generic_message',
                'description': '[{notification.actor}]({notification.actor_link})',
            }
        )
        self._create_notification()
        n = notification_queryset.first()
        self.assertEqual(n.level, 'info')
        self.assertEqual(n.verb, 'generic verb')
        expected_output = (
            '<p><a href="https://example.com{user_path}">admin</a></p>'
        ).format(
            user_path=reverse('admin:openwisp_users_user_change', args=[self.admin.pk])
        )
        self.assertEqual(n.message, expected_output)
        self.assertEqual(n.rendered_description, expected_output)
        self.assertEqual(n.email_subject, '[example.com] Generic Notification Subject')

    def test_notification_level_kwarg_precedence(self):
        # Create a notification with level kwarg set to 'warning'
        self.notification_options.update({'level': 'warning'})
        self._create_notification()
        n = notification_queryset.first()
        self.assertEqual(n.level, 'warning')

    @mock_notification_types
    def test_misc_notification_type_validation(self):
        with self.subTest('Registering with incomplete notification configuration.'):
            with self.assertRaises(AssertionError):
                register_notification_type('test_type', dict())

        with self.subTest('Registering with improper notification type name'):
            with self.assertRaises(ImproperlyConfigured):
                register_notification_type(['test_type'], dict())

        with self.subTest('Registering with improper notification configuration'):
            with self.assertRaises(ImproperlyConfigured):
                register_notification_type('test_type', tuple())

        with self.subTest('Unregistering with improper notification type name'):
            with self.assertRaises(ImproperlyConfigured):
                unregister_notification_type(dict())

    @mock_notification_types
    def test_notification_type_message_template(self):
        message_template = {
            'level': 'info',
            'verb': 'message template verb',
            'verbose_name': 'Message Template Type',
            'email_subject': '[{site.name}] Messsage Template Subject',
        }

        with self.subTest('Register type with non existent message template'):
            with self.assertRaises(TemplateDoesNotExist):
                message_template.update({'message_template': 'wrong/path.md'})
                register_notification_type('message_template', message_template)

        with self.subTest('Registering type with message template'):
            message_template.update(
                {'message_template': 'openwisp_notifications/default_message.md'}
            )
            register_notification_type('message_template', message_template)
            self.notification_options.update({'type': 'message_template'})
            self._create_notification()
            n = notification_queryset.first()
            self.assertEqual(n.type, 'message_template')
            self.assertEqual(n.email_subject, '[example.com] Messsage Template Subject')

        with self.subTest('Links in notification message'):
            url = _get_absolute_url(
                reverse('admin:openwisp_users_user_change', args=(self.admin.pk,))
            )
            message = (
                '<p>info : None message template verb </p>\n'
                f'<p><a href="{url}">admin</a>'
                '\nreports\n<a href="#">None</a>\nmessage template verb.</p>'
            )
            self.assertEqual(n.message, message)

    @mock_notification_types
    def test_register_unregister_notification_type(self):
        from openwisp_notifications.types import NOTIFICATION_CHOICES

        test_type = {
            'verbose_name': 'Test Notification Type',
            'level': 'test',
            'verb': 'testing',
            'message': '{notification.verb} initiated by {notification.actor} since {notification}',
            'email_subject': '[{site.name}] {notification.verb} reported by {notification.actor}',
        }

        with self.subTest('Registering new notification type'):
            register_notification_type('test_type', test_type)
            self.notification_options.update({'type': 'test_type'})
            self._create_notification()
            n = notification_queryset.first()
            self.assertEqual(n.level, 'test')
            self.assertEqual(n.verb, 'testing')
            self.assertEqual(
                n.message,
                '<p>testing initiated by admin since 0\xa0minutes</p>',
            )
            self.assertEqual(n.email_subject, '[example.com] testing reported by admin')

        with self.subTest('Re-registering a notification type'):
            with self.assertRaises(ImproperlyConfigured):
                register_notification_type('test_type', test_type)

        with self.subTest('Check registration in NOTIFICATION_CHOICES'):
            notification_choice = NOTIFICATION_CHOICES[-1]
            self.assertTupleEqual(
                notification_choice, ('test_type', 'Test Notification Type')
            )

        with self.subTest('Unregistering a notification type which does not exists'):
            with self.assertRaises(ImproperlyConfigured):
                unregister_notification_type('wrong type')

        with self.subTest('Unregistering a notification choice which does not exists'):
            with self.assertRaises(ImproperlyConfigured):
                _unregister_notification_choice('wrong type')

        with self.subTest('Unregistering "test_type"'):
            unregister_notification_type('test_type')
            with self.assertRaises(NotificationRenderException):
                get_notification_configuration('test_type')

        with self.subTest('Using non existing notification type for new notification'):
            with patch('logging.Logger.error') as mocked_logger:
                self._create_notification()
                mocked_logger.assert_called_once_with(
                    'Error encountered while creating notification: '
                    'No such Notification Type, test_type'
                )

        with self.subTest('Check unregistration in NOTIFICATION_CHOICES'):
            with self.assertRaises(ImproperlyConfigured):
                _unregister_notification_choice('test_type')

    def test_notification_email(self):
        self._create_operator()
        self.notification_options.update({'type': 'default'})
        self._create_notification()
        self.assertEqual(len(mail.outbox), 1)

    @mock_notification_types
    def test_missing_relation_object(self):
        test_type = {
            'verbose_name': 'Test Notification Type',
            'level': 'info',
            'verb': 'testing',
            'message': (
                '{notification.verb} initiated by'
                '[{notification.actor}]({notification.actor_link}) with'
                ' [{notification.action_object}]({notification.action_link}) for'
                ' [{notification.target}]({notification.target_link}).'
            ),
            'email_subject': (
                '[{site.name}] {notification.verb} reported by'
                ' {notification.actor} with {notification.action_object} for {notification.target}'
            ),
        }
        register_notification_type('test_type', test_type, models=[User])
        self.notification_options.pop('email_subject')
        self.notification_options.update({'type': 'test_type'})

        with self.subTest("Missing target object after creation"):
            operator = self._get_operator()
            self.notification_options.update({'target': operator})
            self._create_notification()
            operator.delete()

            n_count = notification_queryset.count()
            self.assertEqual(n_count, 0)

        with self.subTest("Missing action object after creation"):
            operator = self._get_operator()
            self.notification_options.pop('target')
            self.notification_options.update({'action_object': operator})
            self._create_notification()
            operator.delete()

            n_count = notification_queryset.count()
            self.assertEqual(n_count, 0)

        with self.subTest("Missing actor object after creation"):
            operator = self._get_operator()
            self.notification_options.pop('action_object')
            self.notification_options.pop('url')
            self.notification_options.update({'sender': operator})
            self._create_notification()
            operator.delete()

            n_count = notification_queryset.count()
            self.assertEqual(n_count, 0)

    @mock_notification_types
    def test_notification_type_related_object_url(self):
        test_type = {
            'verbose_name': 'Test Notification Type',
            'level': 'test',
            'verb': 'testing',
            'message': '{notification.verb} initiated by {notification.actor} since {notification}',
            'email_subject': '[{site.name}] {notification.verb} reported by {notification.actor}',
        }
        self.notification_options.update(
            {
                'type': 'test_type',
                'sender': self._create_user(
                    username='actor', email='actor@example.com'
                ),
                'action_object': self._create_user(username='action-object'),
                'target': self._create_user(
                    username='target', email='target@example.com'
                ),
            }
        )
        # Creating notification will fail without registering
        # notification type here
        register_notification_type('test_type', test_type, models=[User])
        self._create_notification()

        with self.subTest('Test related object static URL'):
            # Update the notification type configuration
            unregister_notification_type('test_type')
            test_type['action_object_link'] = 'https://action-object.example.com'
            test_type['actor_link'] = 'https://actor.example.com'
            test_type['target_link'] = 'https://target.example.com'
            register_notification_type('test_type', test_type, models=[User])

            notification = Notification.objects.first()
            self.assertEqual(
                notification.action_url, 'https://action-object.example.com'
            )
            self.assertEqual(notification.actor_url, 'https://actor.example.com')
            self.assertEqual(notification.target_url, 'https://target.example.com')

        with self.subTest('Test related object callable URL'):
            # Update the notification type configuration
            unregister_notification_type('test_type')
            url_generator = 'openwisp_notifications.tests.test_helpers.notification_related_object_url'
            test_type['action_object_link'] = url_generator
            test_type['actor_link'] = url_generator
            test_type['target_link'] = url_generator
            register_notification_type('test_type', test_type, models=[User])

            notification = Notification.objects.first()
            self.assertEqual(
                notification.action_url, 'https://action-object.example.com'
            )
            self.assertEqual(notification.actor_url, 'https://actor.example.com')
            self.assertEqual(notification.target_url, 'https://target.example.com')

    @capture_any_output()
    @mock_notification_types
    @patch('openwisp_notifications.tasks.delete_notification.delay')
    def test_notification_invalid_message_attribute(self, mocked_task):
        self.notification_options.update({'type': 'test_type'})
        test_type = {
            'verbose_name': 'Test Notification Type',
            'level': 'info',
            'verb': 'testing',
            'message': '{notification.actor.random}',
            'email_subject': '[{site.name}] {notification.actor.random}',
        }
        register_notification_type('test_type', test_type)
        self._create_notification()
        notification = notification_queryset.first()
        with self.assertRaises(NotificationRenderException) as context_manager:
            notification.message
        self.assertEqual(
            str(context_manager.exception),
            'Error encountered in rendering notification message',
        )
        with self.assertRaises(NotificationRenderException) as context_manager:
            notification.email_subject
        self.assertEqual(
            str(context_manager.exception),
            'Error encountered in generating notification email',
        )
        mocked_task.assert_called_with(notification_id=notification.id)

    def test_related_objects_database_query(self):
        operator = self._get_operator()
        self.notification_options.update(
            {'action_object': operator, 'target': operator}
        )
        self._create_notification()
        with self.assertNumQueries(1):
            # 1 query since all related objects are cached
            # when rendering the notification
            n = notification_queryset.first()
            self.assertEqual(n.actor, self.admin)
            self.assertEqual(n.action_object, operator)
            self.assertEqual(n.target, operator)

    @patch.object(app_settings, 'CACHE_TIMEOUT', 0)
    def test_notification_cache_timeout(self):
        # Timeout=0 means value is not cached
        operator = self._get_operator()
        self.notification_options.update(
            {'action_object': operator, 'target': operator}
        )
        self._create_notification()

        n = notification_queryset.first()
        with self.assertNumQueries(3):
            # Expect database query for each operation, nothing is cached
            self.assertEqual(n.actor, self.admin)
            self.assertEqual(n.action_object, operator)
            self.assertEqual(n.target, operator)

        # Test cache is not set
        self.assertIsNone(cache.get(Notification._cache_key(self.admin.pk)))
        self.assertIsNone(cache.get(Notification._cache_key(operator.pk)))

    def test_notification_target_content_type_deleted(self):
        operator = self._get_operator()
        self.notification_options.update(
            {'action_object': operator, 'target': operator, 'type': 'default'}
        )
        self._create_notification()
        ContentType.objects.get_for_model(operator._meta.model).delete()
        ContentType.objects.clear_cache()
        # DoesNotExists exception should not be raised.
        operator.delete()

    def test_delete_old_notification(self):
        days_old = 91
        # Create notification with current timestamp
        self.notification_options.update({'type': 'default'})
        self._create_notification()
        # Create notification with older timestamp
        self.notification_options.update(
            {'timestamp': timezone.now() - timedelta(days=days_old)}
        )
        self._create_notification()

        self.assertEqual(notification_queryset.count(), 2)
        tasks.delete_old_notifications.delay(days_old)
        self.assertEqual(notification_queryset.count(), 1)

    @mock_notification_types
    def test_unregistered_notification_type_related_notification(self):
        # Notifications related to notification type should
        # get deleted on unregistration of notification type
        self.notification_options.update({'type': 'default'})
        unregister_notification_type('default')
        self.assertEqual(notification_queryset.count(), 0)

    @mock_notification_types
    def test_notification_type_email_notification_setting_true(self):
        test_type = {
            'verbose_name': 'Test Notification Type',
            'level': 'info',
            'verb': 'testing',
            'message': 'Test message',
            'email_subject': 'Test Email Subject',
            'email_notification': True,
        }

        register_notification_type('test_type', test_type)
        target_obj = self._get_org_user()
        self.notification_options.update({'type': 'test_type', 'target': target_obj})

        with self.subTest('Test user email preference not defined'):
            self._create_notification()
            self.assertEqual(len(mail.outbox), 1)
            self.assertIsNotNone(mail.outbox.pop())

        with self.subTest('Test user email preference is "False"'):
            NotificationSetting.objects.filter(
                user=self.admin,
                type='test_type',
            ).update(email=False)
            self._create_notification()
            self.assertEqual(len(mail.outbox), 0)

    @mock_notification_types
    def test_notification_type_email_notification_setting_false(self):
        test_type = {
            'verbose_name': 'Test Notification Type',
            'level': 'info',
            'verb': 'testing',
            'message': 'Test message',
            'email_subject': 'Test Email Subject',
            'email_notification': False,
        }

        register_notification_type('test_type', test_type)
        target_obj = self._get_org_user()
        self.notification_options.update({'type': 'test_type', 'target': target_obj})

        with self.subTest('Test user email preference not defined'):
            self._create_notification()
            self.assertEqual(len(mail.outbox), 0)

        with self.subTest('Test user email preference is "True"'):
            NotificationSetting.objects.filter(
                user=self.admin,
                type='test_type',
            ).update(email=True)
            self._create_notification()
            self.assertEqual(len(mail.outbox), 1)

    @mock_notification_types
    def test_notification_type_web_notification_setting_true(self):
        self.notification_options.update({'target': self._get_org_user()})
        test_type = {
            'verbose_name': 'Test Notification Type',
            'level': 'info',
            'verb': 'testing',
            'message': 'Test message',
            'email_subject': 'Test Email Subject',
            'web_notification': True,
        }

        register_notification_type('test_type', test_type)
        self.notification_options.update({'type': 'test_type'})

        with self.subTest('Test user web preference not defined'):
            self._create_notification()
            self.assertEqual(notification_queryset.delete()[0], 1)

        with self.subTest('Test user web preference is "False"'):
            NotificationSetting.objects.filter(
                user=self.admin, type='test_type'
            ).update(web=False)
            self._create_notification()
            self.assertEqual(notification_queryset.count(), 0)

    @mock_notification_types
    def test_notification_type_web_notification_setting_false(self):
        target_obj = self._get_org_user()
        self.notification_options.update({'target': target_obj})
        test_type = {
            'verbose_name': 'Test Notification Type',
            'level': 'info',
            'verb': 'testing',
            'message': 'Test message',
            'email_subject': 'Test Email Subject',
            'web_notification': False,
        }

        register_notification_type('test_type', test_type)
        self.notification_options.update({'type': 'test_type'})

        with self.subTest('Test user web preference not defined'):
            self._create_notification()
            self.assertEqual(notification_queryset.count(), 0)

        with self.subTest('Test user email preference is "True"'):
            notification_setting = NotificationSetting.objects.get(
                user=self.admin, type='test_type', organization=target_obj.organization
            )
            notification_setting.email = True
            notification_setting.save()
            notification_setting.refresh_from_db()
            self.assertFalse(notification_setting.email)

        with self.subTest('Test user web preference is "True"'):
            NotificationSetting.objects.filter(
                user=self.admin, type='test_type'
            ).update(web=True)
            self._create_notification()
            self.assertEqual(notification_queryset.count(), 1)

    @mock_notification_types
    def test_notification_type_email_web_notification_defaults(self):
        test_type = {
            'verbose_name': 'Test Notification Type',
            'level': 'info',
            'verb': 'testing',
            'message': 'Test message',
            'email_subject': 'Test Email Subject',
        }
        register_notification_type('test_type', test_type)

        notification_type_config = get_notification_configuration('test_type')
        self.assertTrue(notification_type_config['web_notification'])
        self.assertTrue(notification_type_config['email_notification'])

    def test_inactive_user_not_receive_notification(self):
        target = self._get_org_user()
        self.notification_options.update({'target': target})

        with self.subTest('Test superuser is inactive'):
            self.admin.is_active = False
            self.admin.save()

            self._create_notification()
            self.assertEqual(notification_queryset.count(), 0)

        # Create org admin
        org_admin = self._create_org_user(user=self._get_operator(), is_admin=True)

        with self.subTest('Test superuser is inactive but org admin is active'):
            self._create_notification()
            self.assertEqual(notification_queryset.count(), 1)
            notification = notification_queryset.first()
            self.assertEqual(notification.recipient, org_admin.user)

        # Cleanup
        notification_queryset.delete()

        with self.subTest('Test both superuser and org admin is inactive'):
            org_admin.user.is_active = False
            org_admin.user.save()

            self._create_notification()
            self.assertEqual(notification_queryset.count(), 0)

        with self.subTest('Test superuser is active and org admin is inactive'):
            self.admin.is_active = True
            self.admin.save()

            self._create_notification()
            self.assertEqual(notification_queryset.count(), 1)
            notification = notification_queryset.first()
            self.assertEqual(notification.recipient, self.admin)

    def test_notification_received_only_by_org_admin(self):
        self.admin.delete()
        org_object = self._get_org_user()
        self.notification_options.update({'sender': org_object, 'target': org_object})
        self._create_org_user(
            user=self._create_user(username='user', email='user@user.com')
        )
        org_admin = self._create_org_user(user=self._get_operator(), is_admin=True)

        self._create_notification()
        self.assertEqual(notification_queryset.count(), 1)
        notification = notification_queryset.first()
        self.assertEqual(notification.recipient, org_admin.user)

    @patch(
        'openwisp_notifications.tasks.ns_register_unregister_notification_type.delay',
        side_effect=OperationalError,
    )
    @patch('logging.Logger.warning')
    def test_post_migrate_handler_celery_broker_unreachable(self, mocked_logger, *args):
        post_migrate.send(
            sender=NotificationAppConfig, app_config=NotificationAppConfig
        )
        mocked_logger.assert_called_once()

    @mock_notification_types
    @patch.object(post_migrate, 'receivers', [])
    @patch(
        'openwisp_notifications.tasks.ns_register_unregister_notification_type.delay',
    )
    def test_post_migrate_populate_notification_settings(self, mocked_task, *args):
        with patch.object(app_settings, 'POPULATE_PREFERENCES_ON_MIGRATE', False):
            NotificationAppConfig.ready()
            post_migrate.send(
                sender=NotificationAppConfig, app_config=NotificationAppConfig
            )
            mocked_task.assert_not_called()
        with patch.object(app_settings, 'POPULATE_PREFERENCES_ON_MIGRATE', True):
            NotificationAppConfig.ready()
            post_migrate.send(
                sender=NotificationAppConfig, app_config=NotificationAppConfig
            )
            mocked_task.assert_called_once()

    @patch('openwisp_notifications.types.NOTIFICATION_ASSOCIATED_MODELS', set())
    @patch('openwisp_notifications.tasks.delete_obsolete_objects.delay')
    def test_delete_obsolete_tasks(self, mocked_task, *args):
        user = self._create_user()
        user.delete()
        mocked_task.assert_not_called()

    def test_email_notif_without_notif_setting(self):
        target_obj = self._get_org_user()
        data = dict(
            email_subject='Test Email subject', url='https://localhost:8000/admin'
        )
        self.admin.notificationsetting_set.all().delete()
        Notification.objects.create(
            actor=self.admin,
            recipient=self.admin,
            description='Test Notification Description',
            verb='Test Notification',
            action_object=target_obj,
            target=target_obj,
            data=data,
            type="default",
        )
        self.assertEqual(len(mail.outbox), 0)

    def test_notification_for_unverified_email(self):
        EmailAddress.objects.filter(user=self.admin).update(verified=False)
        self._create_notification()
        # we don't send emails to unverified email addresses
        self.assertEqual(len(mail.outbox), 0)

    def test_that_the_notification_is_only_sent_once_to_the_user(self):
        first_org = self._create_org()
        first_org.organization_id = first_org.id
        second_org = self._create_org(name='second-org')
        second_org.organization_id = second_org.id
        OrganizationUser.objects.create(user=self.admin, organization=first_org)
        OrganizationUser.objects.create(user=self.admin, organization=second_org)
        self.notification_options.update(
            {
                'type': 'default',
                'sender': first_org,
                'target': first_org,
            }
        )
        self._create_notification()
        self.assertEqual(notification_queryset.count(), 1)


class TestTransactionNotifications(TestOrganizationMixin, TransactionTestCase):
    def setUp(self):
        self.admin = self._create_admin()
        self.notification_options = dict(
            sender=self.admin,
            description='Test Notification',
            level='info',
            verb='Test Notification',
            email_subject='Test Email subject',
            url='https://localhost:8000/admin',
        )

    def _create_notification(self):
        return notify.send(**self.notification_options)

    def test_notification_cache_update(self):
        operator = self._get_operator()
        register_notification_cache_update(
            User, post_save, 'operator_name_changed_invalidation'
        )
        self.notification_options.update(
            {'action_object': operator, 'target': operator, 'type': 'default'}
        )
        self._create_notification()
        content_type = ContentType.objects.get_for_model(operator._meta.model)
        cache_key = Notification._cache_key(content_type.id, operator.id)
        operator_cache = cache.get(cache_key, None)
        self.assertEqual(operator_cache.username, operator.username)
        operator.username = 'new operator name'
        operator.save()
        notification = Notification.objects.get(target_content_type=content_type)
        cache_key = Notification._cache_key(content_type.id, operator.id)
        self._create_notification()
        operator_cache = cache.get(cache_key, None)
        self.assertEqual(notification.target.username, 'new operator name')
        # Done for populating cache
        self.assertEqual(operator_cache.username, 'new operator name')



from django.conf import settings
from django.contrib.admin import ModelAdmin
from django.core import checks
from django.utils.module_loading import import_string

from openwisp_notifications import settings as app_settings


@checks.register
def check_cors_configuration(app_configs, **kwargs):
    errors = []
    if not app_settings.HOST:
        return errors

    if not (
        'corsheaders' in settings.INSTALLED_APPS
        and 'corsheaders.middleware.CorsMiddleware' in settings.MIDDLEWARE
    ):
        errors.append(
            checks.Warning(
                msg='Improperly Configured',
                hint=(
                    '"django-cors-headers" is either not installed or improperly configured.'
                    ' CORS configuration is required for using "OPENWISP_NOTIFICATIONS_HOST" settings.'
                    ' Configure equivalent CORS rules on your server if you are not using'
                    ' "django-cors-headers".'
                ),
                obj='Settings',
            )
        )
    return errors


@checks.register
def check_ow_object_notification_widget_setting(app_configs, **kwargs):
    errors = []
    if not isinstance(app_settings.IGNORE_ENABLED_ADMIN, list):
        errors.append(
            checks.Warning(
                msg='Improperly Configured',
                hint=(
                    '"OPENWISP_NOTIFICATIONS_IGNORE_ENABLED_ADMIN" should be a list, '
                    f'{type(app_settings.IGNORE_ENABLED_ADMIN)} provided'
                ),
                obj='Settings',
            )
        )
        return errors

    # Check individual entries of IGNORE_ENABLED_ADMIN
    for path in app_settings.IGNORE_ENABLED_ADMIN:
        if not isinstance(path, str):
            errors.append(
                checks.Error(
                    msg='Improperly Configured',
                    hint=(
                        '"OPENWISP_NOTIFICATIONS_IGNORE_ENABLED_ADMIN" should contain '
                        f'dotted path string to ModelAdmin, found {type(path)}'
                    ),
                    obj='Settings',
                )
            )
            continue
        # Check whether dotted path points to subclass of ModelAdmin class
        try:
            model_admin_cls = import_string(path)
            assert issubclass(model_admin_cls, ModelAdmin)
        except ImportError:
            errors.append(
                checks.Error(
                    msg='Improperly Configured',
                    hint=(
                        f'Failed to import "{path}" defined in '
                        '"OPENWISP_NOTIFICATIONS_IGNORE_ENABLED_ADMIN". '
                        'Make sure you have provided a valid dotted path.'
                    ),
                    obj='Settings',
                )
            )
        except AssertionError:
            errors.append(
                checks.Error(
                    msg='Improperly Configured',
                    hint=(
                        f'"{path}" does not subclasses '
                        '"django.contrib.admin.ModelAdmin". Only derivatives '
                        'ModelAdmin can be added in "OPENWISP_NOTIFICATIONS_IGNORE_ENABLED_ADMIN".'
                    ),
                    obj='Settings',
                )
            )
    return errors



from django.contrib.auth.management import create_permissions
from django.contrib.auth.models import Permission
from django.db import migrations

from openwisp_notifications.migrations import get_swapped_model
from openwisp_users.migrations import (
    create_default_groups as base_create_default_groups,
)


def add_default_permissions(apps, schema_editor):
    group = get_swapped_model(apps, 'openwisp_users', 'Group')

    # To populate all the permissions
    for app_config in apps.get_app_configs():
        app_config.models_module = True
        create_permissions(app_config, apps=apps, verbosity=0)
        app_config.models_module = None

    operator = group.objects.filter(name='Operator')
    operator = operator.first()

    admin = group.objects.filter(name='Administrator')
    admin = admin.first()

    permissions = [
        Permission.objects.get(
            content_type__app_label='openwisp_notifications',
            codename='add_notification',
        ).pk,
        Permission.objects.get(
            content_type__app_label='openwisp_notifications',
            codename='change_notification',
        ).pk,
        Permission.objects.get(
            content_type__app_label='openwisp_notifications',
            codename='delete_notification',
        ).pk,
    ]
    permissions += operator.permissions.all()
    operator.permissions.set(permissions)

    permissions += admin.permissions.all()
    admin.permissions.set(permissions)


class Migration(migrations.Migration):
    dependencies = [
        ('openwisp_notifications', '0001_initial'),
    ]

    operations = [
        migrations.RunPython(
            base_create_default_groups, reverse_code=migrations.RunPython.noop
        ),
        migrations.RunPython(
            add_default_permissions, reverse_code=migrations.RunPython.noop
        ),
    ]
